#include "simparam.h"
#include "ftr.h"
#include "swparam.h"
#include <stdlib.h>
#include <iostream>
#include "aux.h"
using namespace std;

// Copies the data, ID, debug
// Restores the via-points and num_via to original
void copy_ftr(ftr_t* oldf, ftr_t** newf) {
   int num_via = oldf->get_no_via();

   *newf = new ftr_t(oldf->ID, num_via, oldf->org_viax(0), oldf->org_viay(0), oldf->debug);
   (*newf)->data = oldf->data;
   for(int k=1; k<num_via; k++) {
      (*newf)->via_x[k] = oldf->org_viax(k);
      (*newf)->via_y[k] = oldf->org_viay(k);
   }
}


void print_pop(ftr_t** ftr) {
   for(int i=0; i<POP_SIZE; i++) {
//      cout << i << ": "; // DJ
      cout << ftr[i]-> ID << " " << ftr[i]->data << ": " << ftr[i]->no_via << ": ";

      for(int j=0; j<ftr[i]->get_no_via(); j++) {
         cout << "(" << ftr[i]->org_viax(j) << ", " << ftr[i]->org_viay(j) << ") ";
      }

      cout << endl;
   }

   cout << endl << "----------------------------" << endl;
}

void init_pop(ftr_t *ftr[POP_SIZE], float init_x, float init_y, float fin_x, float fin_y, int ID_offset)
{
   for(int i=0; i<POP_SIZE; i++)
   {
      ftr[i] = new ftr_t(ID_offset+i, INIT_NUM_VIA, init_x, init_y, false); // DJ: Change num of via?
      ftr[i]->via_x[INIT_NUM_VIA -1] = fin_x;
      ftr[i]->via_y[INIT_NUM_VIA -1] = fin_y;

      for(int j=1; j<INIT_NUM_VIA -1; j++)
      {
         ftr[i]->via_x[j] = ((float) rand()) * ( XLEVELS*XSIZE / (float) RAND_MAX);
         ftr[i]->via_y[j] = ((float) rand()) * ( YLEVELS*YSIZE / (float) RAND_MAX);
      }
   }

//   cout << "Initial Population: " << endl;
//   print_pop(ftr);
}

void mutate(ftr_t* ftr) {
   if((rand() < MUT_PROB * RAND_MAX)) { 
      int pos = 1+ (rand() % (ftr->get_no_via()-2));

      ftr->via_x[pos] += (rand()/(float) RAND_MAX);
      ftr->via_y[pos] += (rand()/(float) RAND_MAX);
      ftr->via_x[pos] -= 0.5;
      ftr->via_y[pos] -= 0.5;
      if(ftr->via_x[pos] < 0) ftr->via_x[pos] = 0;
      if(ftr->via_y[pos] < 0) ftr->via_y[pos] = 0;
      if(ftr->via_x[pos] >= XLEVELS*XSIZE) ftr->via_x[pos] = XLEVELS*XSIZE;
      if(ftr->via_y[pos] >= YLEVELS*YSIZE) ftr->via_y[pos] = YLEVELS*YSIZE;
   }
}

void insert(ftr_t* ftr) {

// Insert a random via-point. Otherwise the selection is insanely biased towards reducing num of via points
// One extra slot is available at the end of the via list of the input ftr for the insertion
// no_via is already aligned to the expected output
   float len=0, len_targ;

   for(int i=1; i< ftr->no_via-1; i++) {
     len += dist(ftr->via_x[i-1], ftr->via_y[i-1], ftr->via_x[i], ftr->via_y[i]);
   }

   len_targ = len * (rand()/(float) RAND_MAX);

   for(int i=1; i< ftr->no_via-1; i++) {
      float temp;
      temp = dist(ftr->via_x[i-1], ftr->via_y[i-1], ftr->via_x[i], ftr->via_y[i]);
      if(len_targ < temp) { // found the place to insert
         for(int j = ftr->no_via-1; j>i; j--) {
            ftr->via_x[j] = ftr->via_x[j-1];
            ftr->via_y[j] = ftr->via_y[j-1];
         }

         ftr->via_x[i] = ((len_targ/temp)*(ftr->via_x[i] - ftr->via_x[i-1])) + ftr->via_x[i-1];
         ftr->via_y[i] = ((len_targ/temp)*(ftr->via_y[i] - ftr->via_y[i-1])) + ftr->via_y[i-1];

         return;
      }
      else
         len_targ -= temp;
   }
}

// Returns the number of completed packets sorted
int sort(ftr_t** ftr) {  // Selection sort
   ftr_t* temp;
   int loc_min;
   float cost_min;
   int flag = true;


   for(int i=0; i< POP_SIZE; i++) {
      loc_min = -1;
      cost_min = INFINITY;
      flag = false;

      for(int j=i; j< POP_SIZE; j++) {
         if(ftr[j]->no_via == 0) {  // Check for ejected packets
            flag = true;

            if(ftr[j]->data < cost_min) {
               loc_min = j;
               cost_min = ftr[j]->data;
            }
         }
      }

      if(flag == false) // There is no completed packet left to be sorted, including the present location: ftr[i]
         return i;

//         return (POP_SIZE-i);

      temp = ftr[i];
      ftr[i] = ftr[loc_min];
      ftr[loc_min] = temp;
   }

   return POP_SIZE;
}


int apply_g_ops(ftr_t** old_gen, int ID_offset, bool migrate, ftr_t** migr_cand) {
   int idx;
   unsigned int cum_fitness[POP_SIZE+1];
   int size_old_pop;
   ftr_t *new_gen[POP_SIZE];
   ftr_t *migr_org[MIGR_NUM]; // = migr_cand;

   size_old_pop = sort(old_gen);  // Returns the number of packets from which to derive the new generation
//cout << size_old_pop << endl;
//print_pop(old_gen);

   if(migrate) {
      for(int i=0; i<MIGR_NUM; i++)
         migr_org[i] = migr_cand[i];
   }

   for(int i=0; i<MIGR_NUM; i++) {
      if(i < size_old_pop) {
         copy_ftr(old_gen[i], &migr_cand[i]);
         migr_cand[i]->no_via = 0;
      }
      else migr_cand[i] = NULL;
   }

   if(size_old_pop < SUCC_PERC*POP_SIZE) {
      return size_old_pop;
   }

// Migration - candidates are sorted
   if(migrate) {
   int num_via;
   int i, j, k;

// DJ: Can be further optimized - it is enough to make only one pass of old_gen - start with the worst of migr_org
      for(i=0; i < MIGR_NUM; i++) {
         if(migr_org[i] == NULL || (migr_org[i]->data > old_gen[size_old_pop-1]->data)) // Remaining migration candidates are no better than available ones. DJ!! Or replace them anyway?
            break;

         delete old_gen[size_old_pop-1];

         for(j=size_old_pop-2; j >= 0; j--) {
            if(old_gen[j]->data > migr_org[i]->data)
               old_gen[j+1] = old_gen[j];
            else
               break;
         }

         j++;
         copy_ftr(migr_org[i], &old_gen[j]);
            old_gen[j]->no_via = 0;  // To mark that the packet has already been evaluated
      }
//print_pop(old_gen);
      for(i=0; i< MIGR_NUM; i++)
         if(migr_org[i] != NULL) delete migr_org[i];

   } // Migration finished

//   print_pop(old_gen);

// Elitism - Best x% go to new gen by default
// Stop actually sending them out. Just have them stored in the pop table/old_gen

   for(idx=0; idx < ELIT_NUM; idx++) {
      old_gen[idx]-> ID = ID_offset+idx; // Elite's ID updated in pop-tab
   }

// Roulette-wheel set-up 
   cum_fitness[0] = 0;
   float scale_fact = 0; // RAND_MAX / (float) (old_gen[POP_SIZE-1]->data - old_gen[0]->data);

   for(int i=0; i<size_old_pop; i++) {
      scale_fact += 1/old_gen[i]->data;
   }

   scale_fact = RAND_MAX / scale_fact;

   for(int i=1; i<size_old_pop+1; i++) {
      cum_fitness[i] = scale_fact/old_gen[i-1]->data + cum_fitness[i-1];
   }
   if(cum_fitness[size_old_pop] < RAND_MAX) cum_fitness[size_old_pop] = RAND_MAX;


// Crossover
   for(; idx < POP_SIZE; idx += 2) {
      ftr_t *mom, *dad;
      int i;
      int wheel_spin;

   // Roulette wheel selection
      wheel_spin = rand();
      for(i=0; cum_fitness[i+1] < wheel_spin; i++);
      mom = old_gen[i];

      wheel_spin = rand();
      for(i=0; cum_fitness[i+1] < wheel_spin; i++);
      dad = old_gen[i];

   // Cut point selection

      int num_via_dad = dad->get_no_via();
      int num_via_mom = mom->get_no_via();  // Children's no of via between 2 and mom+dad-2 if unconstrained
      int min_chr_len, max_chr_len;


// Cross-overs are nearly symmetric
      if(num_via_dad+num_via_mom >= 2*MAX_NUM_VIA-1) {
         max_chr_len = MAX_NUM_VIA;
         min_chr_len = num_via_dad+num_via_mom-MAX_NUM_VIA;
      }
      else if(num_via_dad+num_via_mom <= 7) {
         min_chr_len = 3;
         max_chr_len = num_via_dad+num_via_mom-3;
      }
      else {
         max_chr_len = (num_via_dad+num_via_mom)/2 +1;
         min_chr_len = num_via_dad + num_via_mom - max_chr_len;
      }


      int child0_len = (rand() % (max_chr_len-min_chr_len+1)) + min_chr_len;
      int min_dad_cut, max_dad_cut;
      if(child0_len > num_via_mom) min_dad_cut = child0_len - num_via_mom;
      else min_dad_cut = 0;

      if(child0_len > num_via_dad) max_dad_cut = num_via_dad-2;
      else max_dad_cut = child0_len-2;

      int dad_cut, mom_cut;
      if(max_dad_cut == min_dad_cut) dad_cut = max_dad_cut;
      else dad_cut = min_dad_cut + (rand() % (max_dad_cut-min_dad_cut));
      mom_cut = dad_cut + num_via_mom - child0_len;

      int child1_len = num_via_dad + num_via_mom - child0_len;
      int flag0 = 0, flag1 = 0;

      if(child0_len <= child1_len && child0_len < MAX_NUM_VIA && rand() < INSERT_PROB * RAND_MAX)
         flag0 = 1;
      if(child1_len <= child0_len && child1_len < MAX_NUM_VIA && rand() < INSERT_PROB * RAND_MAX)
         flag1 = 1;

      new_gen[idx]   = new ftr_t(ID_offset+idx,   child0_len+flag0, dad->org_viax(0), dad->org_viay(0), false); // First dad_cut+1 genes from dad
      new_gen[idx+1] = new ftr_t(ID_offset+idx+1, child1_len+flag1, dad->org_viax(0), dad->org_viay(0), false); // First mom_cut+1 genes from mom
// Note that from here on till insert() is called, the no_via is one more if the corresponding flag is set

      for(i=1; i <= dad_cut; i++) {
         new_gen[idx]->via_x[i] = dad->org_viax(i);
         new_gen[idx]->via_y[i] = dad->org_viay(i);
      }

      for(int momi = mom_cut+1; i < child0_len; i++, momi++) {
         new_gen[idx]->via_x[i] = mom->org_viax(momi);
         new_gen[idx]->via_y[i] = mom->org_viay(momi);
      }

      for(i=1; i <= mom_cut; i++) {
         new_gen[idx+1]->via_x[i] = mom->org_viax(i);
         new_gen[idx+1]->via_y[i] = mom->org_viay(i);
      }

      for(int dadi = dad_cut+1; i< child1_len; i++, dadi++) {
         new_gen[idx+1]->via_x[i] = dad->org_viax(dadi);
         new_gen[idx+1]->via_y[i] = dad->org_viay(dadi);
      }

      if(flag0 == 1) insert(new_gen[idx]);
      if(flag1 == 1) insert(new_gen[idx+1]);
// At this point no_via is again aligned to the actual value

// Location Mutation
      mutate(new_gen[idx]);
      mutate(new_gen[idx+1]);
    }

// Stop deleting the elits from pop table
   for(int i=ELIT_NUM; i<POP_SIZE; i++) {
      delete old_gen[i];
      old_gen[i] = new_gen[i];
   }

   return -1;
}

ftr_t *eject_oqueue(ftr_ll **head, ftr_ll **tail, int num_act_packets) {
   int eject_idx;
   ftr_t *ftr;
   ftr_ll *tmp_link, *tmp_par;

   if(*head == NULL) { // Empty queue - error condition
      return NULL;
   }

   eject_idx = (rand() / ((float) RAND_MAX+1)) * num_act_packets; // Random index to be ejected

   if(eject_idx == 0) { // Head to be ejected
      ftr = (*head) -> ftr;
      tmp_link = *head;
      *head = (*head) -> nextp;
      if(*head == NULL)
         *tail = NULL;

      delete tmp_link;

      ftr->no_via = 2;
      return ftr;
   }


   tmp_par = *head;
   eject_idx--;

   for(; eject_idx > 0; eject_idx--) {
      tmp_par = tmp_par->nextp;
   }

   tmp_link = tmp_par->nextp;
   ftr = tmp_link->ftr;
   tmp_par->nextp = tmp_link->nextp;

   if(tmp_link == *tail)
      *tail = tmp_par;

   delete tmp_link;

   ftr->no_via = 2;
   return ftr;

}

ftr_t *read_oqueue(ftr_ll **head, ftr_ll **tail, unsigned long long *clknum, bool flush, float ej_prob) {
   ftr_t *ftr;
   ftr_ll *tmp_link;

   if(*head == NULL) { // Empty queue
      return NULL;
   }

   if(((*head)->clk_num < *clknum) || flush) {
      tmp_link = *head;
      ftr = (*head) -> ftr;

      if((*head)->nextp == NULL) { // Only 1 element in queue
         *head = NULL;
         *tail = NULL;
      } else {
         *head = (*head)->nextp;
      }

      if(flush) *clknum = tmp_link->clk_num;

      delete tmp_link;

      if(rand() < RAND_MAX * ej_prob)
         ftr->no_via = 2;  // Marking the packet unfinished

      return ftr;
   }
   else
      return NULL; // Nothing to be read out NOW, but for later
}

void write_oqueue(ftr_ll **head, ftr_ll **tail, ftr_t *ftr, unsigned long long clknum) {
   ftr_ll *tmp_link;


   tmp_link = new ftr_ll(ftr);
   tmp_link -> clk_num = clknum; // clk_num


   if(*tail == NULL) { // queue empty
      *head = tmp_link;
      *tail = tmp_link;
      return;
   }


   if(clknum < (*head)->clk_num) {  // Insertion at the head
      tmp_link -> nextp = *head;
      *head = tmp_link;
      return;
   }



   if(clknum >= (*tail)->clk_num) { // Insertion at the tail
      (*tail) -> nextp = tmp_link;
      *tail = tmp_link;
      return;
   }


   for(ftr_ll *pres = *head; ; pres = pres->nextp) {  // Insertion between head and tail
      if(pres -> nextp -> clk_num > clknum) {
         tmp_link -> nextp = pres -> nextp;
         pres -> nextp = tmp_link;
         return;
      }
   }

}

void print_queue(ftr_ll* head) {
   while (head != NULL) {
      cout << "ID: " << head->ftr->ID << ", Finish time: " << head->clk_num << endl;
      head = head->nextp;
   }
}

int adapt_disp_rate(int *throughput, bool ejected_pack) {
   static bool pack_status[DISP_ADP_PER];
   static int  pack_stat_h = 0;
   int no_eject;
   const int max_no_eject = MAX_EJ_RATIO * DISP_ADP_PER;
   const int min_no_eject = (MIN_EJ_RATIO > MAX_EJ_RATIO) ? 0 : (MIN_EJ_RATIO * DISP_ADP_PER);
   const int targ_no_eject = (max_no_eject + min_no_eject)/2;

   pack_status[pack_stat_h] = ejected_pack;
   if(pack_stat_h < DISP_ADP_PER-1) {
      pack_stat_h++;
      return -1;
   }
   else {
      pack_stat_h = 0;
      no_eject = 0;
      for(int i=0; i< DISP_ADP_PER; i++) {
         if(pack_status[i]) no_eject++;
      }

      if(no_eject > max_no_eject)
        *throughput *= (1.0 + (RELAX_RATE * (no_eject-targ_no_eject) / (float) (DISP_ADP_PER - targ_no_eject))); // increase throughput

      if(no_eject < min_no_eject)
        *throughput *= (1.0 - (TIGHT_RATE * (targ_no_eject-no_eject) / (float) targ_no_eject));    // decrease throughput

      return no_eject;
   }
}

