#include "simparam.h"
#include "ftr.h"
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <iostream>
#include <fstream>
#include <pthread.h>
#include "swparam.h"
#include "aux.h"
using namespace std;

// DJ: ptab_mx is not needed if always going for full population.
pthread_mutex_t *dispq_mx, *hwop_mx, *ptab_mx;
ftr_ll *dispq_h, *dispq_t;
ftr_ll *hwop_h, *hwop_t;
ftr_t *pop_tab[NUM_POP][POP_SIZE];
int  pop_tab_cnt[NUM_POP];
bool finish;
bool hw_finish;


ftr_t *read_queue(ftr_ll **head, ftr_ll **tail, pthread_mutex_t *mutex) {
   ftr_t *ftr;
   ftr_ll *tmp_link;

   pthread_mutex_lock(mutex);
   ftr = (*head)->ftr;
   tmp_link = *head;

   if((*head)->nextp == NULL) { // Only 1 element in queue
      *head = NULL;
      *tail = NULL;
   } else {
      *head = (*head)->nextp;
   }

   pthread_mutex_unlock(mutex);
   delete tmp_link;
   return ftr;
}

void write_queue(ftr_ll **head, ftr_ll **tail, pthread_mutex_t *mutex, ftr_t *ftr) {
   ftr_ll *tmp_link = new ftr_ll(ftr);

   pthread_mutex_lock(mutex);
   if(*tail == NULL) { // queue empty
      *head = tmp_link;
      *tail = tmp_link;
   } else {
      (*tail) -> nextp = tmp_link;
      *tail = tmp_link;
   }
   pthread_mutex_unlock(mutex);
}

void* Gen_operator (void *dummy) {
   int iter = 0;
   int gen[NUM_POP];
   ftr_t *ftr_tmp;
   int pop = 0;
   ftr_t *migr_cand[MIGR_NUM];
//   int new_migr_num;


   for(int i=0; i<NUM_POP; i++)
      gen[i] = 0;

   for(int i=0; i<MIGR_NUM; i++)
      migr_cand[i] = NULL;

   while(1) {
      if(pop_tab_cnt[pop] == POP_SIZE) {

         if((iter+1)%20 == 0 && pop == 0) {
            float best = pop_tab[0][0]->data;  // DJ: Seg fault if elitism = 0
            cout << "Iteration " << iter << ":" << endl;
            for(int i=0; i<NUM_POP; i++) {
               cout << "Pop, gen " << i << ", " << gen[i] << ": " << pop_tab[i][0]->data << endl;  // DJ: Seg fault if elitism = 0
               if(best > pop_tab[i][0]->data) best = pop_tab[i][0]->data;
            }
            cout << "Best = " << best << endl << endl;

         }

         if(iter == NUM_GEN) {
            break;
         }

         int migr_pop = (pop == NUM_POP-1) ? 0 : pop+1;

 // Apply genetic operations on pop_tab and build new generation in situ
//cout << "Pop: " << pop << ", Gen: " << gen[pop] << ", Iter: " << iter << endl;
         int num_success = apply_g_ops(pop_tab[pop], pop*POP_SIZE, (gen[pop]%MIGR_GEN_GAP == 0) && gen[pop] != 0 && MIGR_ENABLE, migr_cand);

/*
         if(num_success >= 0 && num_success < MIGR_NUM) new_migr_num = num_success;
         else new_migr_num = MIGR_NUM;

         for(new_migr_num--; new_migr_num >= 0; new_migr_num--) {
            delete migr_cand[new_migr_num];
            copy_ftr(pop_tab[pop][new_migr_num], &migr_cand[new_migr_num]);
         } */

         if(num_success == -1) {    // New population available

            pthread_mutex_lock(ptab_mx);
               pop_tab_cnt[pop] = ELIT_NUM;
            pthread_mutex_unlock(ptab_mx);

            gen[pop]++; // DJ: won't happen when there are lost packets

// Elit populations are not sent out
            for(int i=ELIT_NUM; i<POP_SIZE; i++) {
               copy_ftr(pop_tab[pop][i], &ftr_tmp);
               write_queue(&dispq_h, &dispq_t, dispq_mx, ftr_tmp);
            }
         } else {  // Need to resend the packets for finishing computations

            pthread_mutex_lock(ptab_mx);
               pop_tab_cnt[pop] = num_success;
            pthread_mutex_unlock(ptab_mx);


            for(int i=num_success; i<POP_SIZE; i++) {
               copy_ftr(pop_tab[pop][i], &ftr_tmp); // Restores the packet to pristine, except data.
               ftr_tmp->data = 0;
               ftr_tmp->ID = i+pop*POP_SIZE;
         //      delete pop_tab[pop][i];

               write_queue(&dispq_h, &dispq_t, dispq_mx, ftr_tmp); 
            }

         }

         if(pop == NUM_POP-1) iter++;
         if(pop == NUM_POP-1) pop = 0;  // Populations are expected to finish in sequence.
         else pop++;

         if(iter == NUM_GEN)
            finish = true;


      } else {
         usleep(25);
      }
      
   }
}

void* Gatherer(void *dummy) {
   ftr_t *ftr;

   while(1) {
      if(hwop_h != NULL) { // HW OP available
         ftr = read_queue(&hwop_h, &hwop_t, hwop_mx);
         int pop = (ftr->ID/POP_SIZE)%NUM_POP;
         int idx = (ftr->ID)%POP_SIZE;


         if(pop_tab[pop][idx]->get_no_via() != ftr->get_no_via())
            cout << "Warning!" << endl;

//         pop_tab[pop][pop_tab_cnt[pop]] = ftr; // Pop table is filled up in the order that the packets are received
         pop_tab[pop][idx]-> data = ftr->data;  // ftr is put into the slot implied by its ID - also ejected packets
         pop_tab[pop][idx]-> no_via = ftr->no_via;

         delete ftr;

         pthread_mutex_lock(ptab_mx);
         pop_tab_cnt[pop]++;
         pthread_mutex_unlock(ptab_mx);

      } else if(hw_finish) {
         break;
      } else {
         usleep(25);
      }
   }
}

void* Dispatcher(void *dummy) {
   ftr_t *ftr;
   unsigned long long clk_num = 0;
   unsigned long long next_eject = EJECT_PER;
   unsigned long long next_inject = 0;
   int diff_time;
   ofstream timelog, summary, loadprofile, adaptlog;
   int num_act_packets = 0;
   ftr_ll *orderq_h, *orderq_t;
   unsigned long long finish_clk_[NUM_POP];
//   bool pack_status[DISP_ADP_PER];
//   int  pack_stat_h = 0;
   int  throughput_adp = THROUGHPUT;
   int tmp_flag;


   orderq_h = orderq_t = NULL;
   timelog.open("timing.log");
   summary.open("timing_summary.log");
   loadprofile.open("load_profile.log");
   adaptlog.open("adapt_dispatch.log");
   timelog << "ID\tStart\tFinish\tDifferential" << endl << endl;
   hw_finish = false;
   for(int i=0; i<NUM_POP; i++)
      finish_clk_[i] = 0;

   adaptlog << "0 \t " << THROUGHPUT << endl;

   while(1) {
// End condition
      if(finish && dispq_h == NULL && orderq_h == NULL) {
         timelog.close();
         summary.close();
         loadprofile.close();
         adaptlog.close();
         cout << "HW finishes at clock no: " << clk_num << endl;
         hw_finish = true;
         break;
      }

// Producing HW outputs - normal case
      if(next_eject <= next_inject) {  // It's not yet time to inject
         clk_num = next_eject;         // DJ: What if nothing available to inject? Maybe lack of synch with other threads, or deadlock, or finish.
         next_eject += EJECT_PER;

         while((ftr=read_oqueue(&orderq_h, &orderq_t, &clk_num, false, 0.0)) != NULL) { // Existing packets ready to be sent out: clk_num not changed
            write_queue(&hwop_h, &hwop_t, hwop_mx, ftr);
            num_act_packets--;

            if(DISP_ADP_ENA) {
               tmp_flag = adapt_disp_rate(&throughput_adp, false);
               if(tmp_flag > -1) adaptlog << clk_num  << "\t" << throughput_adp  << "\t" << tmp_flag << endl;
            }
         }

         while((num_act_packets-EJECT_THR0)/ (float) (EJECT_THR1-EJECT_THR0) > rand() / (float) RAND_MAX) { // Keep on ejecting till ejection fails once
            ftr = eject_oqueue(&orderq_h, &orderq_t, num_act_packets);
            num_act_packets--;
            write_queue(&hwop_h, &hwop_t, hwop_mx, ftr);

            if(DISP_ADP_ENA) {
               tmp_flag = adapt_disp_rate(&throughput_adp, true);
               if(tmp_flag > -1) adaptlog << clk_num  << "\t" << throughput_adp  << "\t" << tmp_flag << endl;
            }
         }

         loadprofile << clk_num << "\t" << num_act_packets << endl;

      }

// Reading HW inputs
      else if(dispq_h != NULL) {  // Dispatch queue not empty
         clk_num = next_inject;
         next_inject += throughput_adp;

         ftr = read_queue(&dispq_h, &dispq_t, dispq_mx);
         num_act_packets++;
         timelog << ftr->ID << "\t" << clk_num;

         if(ftr->ID % POP_SIZE == ELIT_NUM) // First packet of population
            summary << "Pop #" << (ftr->ID / POP_SIZE) << ": Start = " << clk_num << endl; // Allow Pop# go beyond num_pop?
         diff_time = traj_cost(ftr);
         timelog << "\t" << clk_num+diff_time << "\t" << diff_time << endl;

         if(clk_num+diff_time > finish_clk_[(ftr->ID / POP_SIZE) % NUM_POP])
            finish_clk_[(ftr->ID/POP_SIZE) % NUM_POP] = clk_num+diff_time;  // Update the worst finish time for the population


         if(((ftr->ID)+1)%POP_SIZE == 0) // Last packet of population
            summary << "Pop #" << (ftr->ID / POP_SIZE) << ": Finish = " << finish_clk_[(ftr->ID/POP_SIZE) % NUM_POP] << endl;


         write_oqueue(&orderq_h, &orderq_t, ftr, clk_num+diff_time);

      } // Added

// Below: corner cases of of HW-modelling.
// Produce HW output - flushing
      else if(finish) {
         while((ftr=read_oqueue(&orderq_h, &orderq_t, &clk_num, true, 0.0)) != NULL)  { // Flush the hardware: clk_num is updated at reading
            write_queue(&hwop_h, &hwop_t, hwop_mx, ftr);
            num_act_packets--;
        //    adapt_disp_rate(&throughput_adp, false, adapt); // Meaningless as finishing simulation, no more injection will happen
         }
      }

      else {
         while (hwop_h != NULL && dispq_h == NULL)
            usleep(30);  // Give chance to gatherer to gather all packets or genetic operator to finish

         if(dispq_h == NULL)
            usleep(40);  // Allow gen op some more time to finish after gatherer finishes.

         if(dispq_h == NULL) { // It is likely a SW simulation deadlock

            if(NUM_POP > 1)  // Simulation deadlocks are supposed to be rare
               cout << "SW simulation deadlock: " << clk_num << endl;

            if((ftr=read_oqueue(&orderq_h, &orderq_t, &clk_num, true, 0.0)) != NULL)  { // Read out next HW packet: clk_num is updated at reading
               write_queue(&hwop_h, &hwop_t, hwop_mx, ftr);
               num_act_packets--;
          //     adapt_disp_rate(&throughput_adp, false, adapt); // Disable adaptation when breaking SW deadlock, i.e. no injection happening.
            }
            if(clk_num > next_eject)
               next_eject = clk_num;
            next_inject = next_eject; // This may allow some packet dropping due to HW deadlock 
         }
      }
   }
}


int main(int argc, char *argv[]) {
   ftr_t* ftr[POP_SIZE];
   int seed;
   pthread_t gatherer, gen_operator, dispatcher;
   ftr_t* pop_table[POP_SIZE];
   ftr_ll *tmp;


   finish = false;
   dispq_h = dispq_t = NULL;
   hwop_h = hwop_t = NULL;
   for(int i=0; i<NUM_POP; i++)
      pop_tab_cnt[i] = 0;

   if(argc > 1) seed = atoi(argv[1]);
   else seed = time(NULL);
   srand(seed);
   cout << "Simulation seed = " << seed << endl << endl;

//   init_pop(ftr, 1.0, 1.4, 17.0, 17.0, 11);
/*   init_pop(ftr, 3, 11, 15, 11, 0); // (3,11) -> (15,11). First packet's ID = 0.
 */


   for(int j = 0; j<NUM_POP; j++) { // Initialize all populations   
      init_pop(ftr, 3, 11, 15, 11, j*POP_SIZE); // Path from (3,11) to (15,11).

      for(int i=0; i<POP_SIZE; i++) {
         tmp = new ftr_ll(ftr[i]);
         copy_ftr(ftr[i], &pop_tab[j][i]);

         if(i == 0 && j==0) { // Empty queue
            dispq_t = dispq_h = tmp;
         } else {
            dispq_t->nextp = tmp;
            dispq_t = tmp;
         }
      }
   }

   dispq_mx = (pthread_mutex_t*) malloc(sizeof(pthread_mutex_t));
   hwop_mx = (pthread_mutex_t*) malloc(sizeof(pthread_mutex_t));
   ptab_mx = (pthread_mutex_t*) malloc(sizeof(pthread_mutex_t));

   pthread_mutex_init(dispq_mx, NULL);
   pthread_mutex_init(hwop_mx, NULL);
   pthread_mutex_init(ptab_mx, NULL);

   pthread_create(&dispatcher, NULL, Dispatcher, NULL);
   pthread_create(&gatherer, NULL, Gatherer, NULL);
   pthread_create(&gen_operator, NULL, Gen_operator, NULL);

//   get_best(ftr_tmp);
//   print_pop(ftr);
   pthread_join(gen_operator, NULL);
   pthread_join(gatherer, NULL);
   pthread_join(dispatcher, NULL);

   cout << endl << "DONE" << endl;
/*
   for(int i=0; i<POP_SIZE; i++) {
      delete ftr[i];
   } */

//   cout << "Final: (" << presx << ", " << presy << ") -> " << pres_cost << endl;

   return 0;
}
