`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date:    22:13:11 05/23/2014 
// Design Name: 
// Module Name:    proc_node_with_mcu 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: 
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
module proc_node_with_mcu(
 input rst,
 input clk,

 input  ie_req_ds,
 input  [IO_W-1:0] ie_pack_ds,
 input  [(COORD_SIZE*2)-1:0] ie_add_ds,
 output ie_grant_ds,
 output ie_reject_ds,
 input  ie_valid_ds, // IENW interface - downstream ports, driven by an upstream proc_node
 
 output ie_req_us,
 output [IO_W-1:0] ie_pack_us,
 output [(COORD_SIZE*2)-1:0] ie_add_us,
 input  ie_grant_us,
 input  ie_reject_us,
 output ie_valid_us, // IENW interface - upstream ports, to drive a downstream proc_node

 input  [3:0] wr_req_ds,
 input  [4*IO_W-1:0] wr_pack_ds,
 output [3:0] wr_grant_ds,
 input  [3:0] wr_valid_ds, // CNW interface - downstream ports, driven by an upstream proc_node
 
 output [3:0] wr_req_us,
 output [4*IO_W-1:0] wr_pack_us,
 input  [3:0] wr_grant_us,
 output [3:0] wr_valid_us // CNW interface - upstream ports, to drive a downstream proc_node
     );

parameter TIMEOUT_W = 10;
parameter WORD_W = 32;    // Width of each word, same as width of the memory interface. Must be even.
parameter MEM_ADD_W = 10;
parameter HEADER_SIZE = 5; // size of ftr packet header in number of words
parameter FIFO_DEPTH_W = 7; // Depth of FIFO in number of words = (2^FIFO_DEPTH_W -1) : -1 for distinguishing empty/full
parameter PN_Y_COORD = 0;
parameter PN_X_COORD = 0;
parameter COORD_SIZE = 8;
parameter NUM_VIA_W = FIFO_DEPTH_W;
parameter REC_LIST_ADD = 5 * (2 ** FIFO_DEPTH_W);
parameter FLOP_ON_IENW_REQ = 0;
parameter FLOP_ON_IENW_GRANT_REJ = 0;
parameter TIMEOUT_COUNT_PREC = 0; // Timeout is counted every 2**TIMEOUT_COUNT_PREC clocks

localparam SEND_LIST_BASE = 2**MEM_ADD_W;
localparam IO_W = WORD_W/2;

 wire  [MEM_ADD_W:0] cpu_add;
 wire   cpu_wr;
 wire   cpu_cs;
 wire [WORD_W-1:0] cpu_data_i;
 wire  [WORD_W-1:0] cpu_data_o; // Processor interface

 wire  add_str, rd_str;
 reg ready;
 
 always @(posedge clk or posedge rst)
    if(rst)
	    ready <= 1'b0;
	 else
	    ready <= cpu_cs;
		 

 
 assign cpu_cs = add_str | rd_str;
microblaze_mcs_v1_2 microblaze_inst(
   .Clk(clk),
	.Reset(rst),
	
	.IO_Addr_Strobe(add_str),
	.IO_Read_Strobe(rd_str),
	.IO_Write_Strobe(cpu_wr),
	.IO_Address(cpu_add),
	.IO_Byte_Enable(),
	.IO_Write_Data(cpu_data_o),
	.IO_Read_Data(cpu_data_i),
	.IO_Ready(ready)
    );


proc_node_with_ie #(
.TIMEOUT_W(TIMEOUT_W),
.WORD_W(WORD_W),    // Width of each word, same as width of the memory interface. Must be even.
.MEM_ADD_W(MEM_ADD_W),
.HEADER_SIZE(HEADER_SIZE), // size of ftr packet header in number of words
.FIFO_DEPTH_W(FIFO_DEPTH_W), // Depth of FIFO in number of words = (2^FIFO_DEPTH_W -1) : -1 for distinguishing empty/full
.PN_Y_COORD(PN_Y_COORD),
.PN_X_COORD(PN_X_COORD),
.COORD_SIZE(COORD_SIZE),
.NUM_VIA_W(NUM_VIA_W),
.REC_LIST_ADD(REC_LIST_ADD),
.FLOP_ON_IENW_REQ(FLOP_ON_IENW_REQ),
.FLOP_ON_IENW_GRANT_REJ(FLOP_ON_IENW_GRANT_REJ),
.TIMEOUT_COUNT_PREC(TIMEOUT_COUNT_PREC) // Timeout is counted every 2**TIMEOUT_COUNT_PREC clocks

) proc_node0(
 .rst(rst),
 .clk(clk),

 .ie_req_ds(ie_req_ds),
 .ie_pack_ds(ie_pack_ds),
 .ie_add_ds(ie_add_ds),
 .ie_grant_ds(ie_grant_ds),
 .ie_reject_ds(ie_reject_ds),
 .ie_valid_ds(ie_valid_ds), // IENW interface - downstream ports, driven by an upstream proc_node
 
 .ie_req_us(ie_req_us),
 .ie_pack_us(ie_pack_us),
 .ie_add_us(ie_add_us),
 .ie_grant_us(ie_grant_us),
 .ie_reject_us(ie_reject_us),
 .ie_valid_us(ie_valid_us), // IENW interface - upstream ports, to drive a downstream proc_node
 
 .wr_req_ds(wr_req_ds),
 .wr_pack_ds(wr_pack_ds),
 .wr_grant_ds(wr_grant_ds),
 .wr_valid_ds(wr_valid_ds), // IENW & CNW interface - downstream ports, driven by an upstream proc_node
 
 .wr_req_us(wr_req_us),
 .wr_pack_us(wr_pack_us),
 .wr_grant_us(wr_grant_us),
 .wr_valid_us(wr_valid_us), // IENW & CNW interface - upstream ports, to drive a downstream proc_node
 
 .cpu_add(cpu_add),
 .cpu_wr(cpu_wr),
 .cpu_cs(cpu_cs),
 .cpu_data_i(cpu_data_i),
 .cpu_data_o(cpu_data_o) // Processor interface
);

endmodule
