module send_mod (
 input rst,
 input clk,
 output [(COORD_SIZE*2) -1:0] our_add,
 
 output reg [4:0] wr_req,
 output reg [5*IO_W-1:0] wr_pack,
 input [4:0] wr_grant,
 output reg [4:0] wr_valid, // IENW & CNW interface

 output rd_req,
 input [IO_W-1:0] rd_pack,
 input   rd_valid,
 output [2:0] rd_fifon,
 output [NUM_VIA_W-1:0] rd_num_via, // FIFO-controller interface

 input [2:0] cpu_add,
 input  cpu_wr,
 input  cpu_cs,
 input [4+TIMEOUT_W+NUM_VIA_W:0] cpu_data_i,
 output reg [4+TIMEOUT_W+NUM_VIA_W:0] cpu_data_o  // Send-list memory-like interface
 // CPU data format: {sent_toggle, ready_toggle, timeout, route_dir, num_via}. sent_toggle ignored when writing into send-list
);

   parameter TIMEOUT_W = 10;
   parameter NUM_VIA_W = 7;
	parameter IO_W = 16;
	parameter PN_Y_COORD = 0;
	parameter PN_X_COORD = 0;
	parameter COORD_SIZE = 8;
	parameter TIMEOUT_COUNT_PREC = 0; // Timeout is counted every 2**TIMEOUT_COUNT_PREC clocks

   reg [4:0] sent_toggle;
   reg [4:0] ready_toggle;
   reg [(5*TIMEOUT_W)-1:0] timeout;
   reg [14:0] route_dir;
   reg [(5*NUM_VIA_W)-1:0] num_via;

   reg [14:0] arb_tab;
   reg [3:0] state;
   reg [3:0] next_state;
   reg  [2:0] idx;
   wire [2:0] pres_route_dir;

   wire smd_finish, smd_reroute; // , smd_decr
	wire [2:0] smd_add;
	
	wire timeout_tick;

   integer temp;
	
	wire [COORD_SIZE-1 :0] pn_y_coord = PN_Y_COORD;
	wire [COORD_SIZE-1 :0] pn_x_coord = PN_X_COORD;
	
	assign our_add = {pn_y_coord, pn_x_coord};

   generate
      if (NUM_VIA_W >= IO_W)
         error_PackSize_gt_IOWidth error_inst0();
		if (COORD_SIZE*2 > 5+TIMEOUT_W+NUM_VIA_W)
		   error_CoordSize_gt_CpuDataSize error_inst1();
   endgenerate
	
	generate
	   if(TIMEOUT_COUNT_PREC == 0)
		   assign timeout_tick = 1'b1;
		else
		begin
		   reg [TIMEOUT_COUNT_PREC-1 :0] timeout_prec;
			
			always @(posedge clk or posedge rst)
			   if(rst)
				   timeout_prec <= -1;
				else
				   timeout_prec <= timeout_prec + 1; // Allowing rollover
					
			assign timeout_tick = (timeout_prec == 0);
		end
	endgenerate
//-------------------------------------------
//
//     Send - list as a Reg bank
//
//--------------------------------------------

   always @(posedge clk or posedge rst)
	   if(rst)
		begin
			sent_toggle <= 5'd0;
		   ready_toggle <= 5'd0;
//			timeout <= {5*TIMEOUT_W{1'b0}};
			route_dir <= {15{1'b1}};     // Invalid direction
			num_via <= {5*NUM_VIA_W{1'b0}};
		end
		else
		begin
		   if(cpu_wr && cpu_cs && ready_toggle[cpu_add] == sent_toggle[cpu_add])
		   begin
		      ready_toggle[cpu_add] <= cpu_data_i[3+NUM_VIA_W+TIMEOUT_W];
//   			timeout[TIMEOUT_W*cpu_add+:TIMEOUT_W] <= cpu_data_i[NUM_VIA_W+2+TIMEOUT_W:NUM_VIA_W+3];
			   route_dir[3*cpu_add+:3] <= cpu_data_i[NUM_VIA_W+2:NUM_VIA_W];
			   num_via[NUM_VIA_W*cpu_add+:NUM_VIA_W] <= cpu_data_i[NUM_VIA_W-1:0];
		   end
		   else if(ready_toggle[smd_add] != sent_toggle[smd_add])
			begin
			   if(smd_finish) sent_toggle[smd_add] <= ~sent_toggle[smd_add];
//				if(smd_decr)   timeout[smd_add*TIMEOUT_W+:TIMEOUT_W] <= timeout[smd_add*TIMEOUT_W+:TIMEOUT_W]-1;
				if(smd_reroute) route_dir[smd_add*3+:3] <= 3'd0;
			end
		end

	always @(posedge clk or posedge rst)
	   if(rst)
		   timeout <= {5*TIMEOUT_W{1'b0}};
		else
		begin
		   for(temp=0; temp < 5; temp = temp +1)
			begin
			   if(ready_toggle[temp] == sent_toggle[temp] && temp == cpu_add && cpu_wr && cpu_cs)
				   timeout[TIMEOUT_W*temp +:TIMEOUT_W] <= cpu_data_i[NUM_VIA_W+2+TIMEOUT_W:NUM_VIA_W+3];
				else if(   timeout_tick == 1'b1
				           && arb_tab[3*route_dir[temp*3 +:3] +:3] == temp // We are requesting output access
				           && timeout[TIMEOUT_W*temp +:TIMEOUT_W] > 0)  //   and not yet timed out
				   timeout[TIMEOUT_W*temp +:TIMEOUT_W] <= timeout[TIMEOUT_W*temp +:TIMEOUT_W] -1;
			end
		end
	
	always @(posedge clk or posedge rst)
		if(rst)
			cpu_data_o <= 0;
		else if(!cpu_wr && cpu_cs)
		begin
		   if(cpu_add < 5)
			begin
		      cpu_data_o[4+NUM_VIA_W+TIMEOUT_W] <= sent_toggle[cpu_add];
			   cpu_data_o[3+NUM_VIA_W+TIMEOUT_W] <= ready_toggle[cpu_add];
			   cpu_data_o[2+NUM_VIA_W+TIMEOUT_W:NUM_VIA_W+3] <= timeout[TIMEOUT_W*cpu_add+:TIMEOUT_W];
			   cpu_data_o[NUM_VIA_W+2:NUM_VIA_W] <= route_dir[3*cpu_add+:3];
			   cpu_data_o[NUM_VIA_W-1:0] <= num_via[NUM_VIA_W*cpu_add+:NUM_VIA_W];
			end
			else if(cpu_add == 5)
			   cpu_data_o <= our_add;
			else
			   cpu_data_o <= 0;
		end
		else
		   cpu_data_o <= 0;


//-------------------------------------------------
//
//   Send-module state machine
//
//-------------------------------------------------

   assign pres_route_dir = route_dir[3*idx+:3];

   always @(posedge clk or posedge rst)
	   if(rst)
		   state <= 4'b0000;
		else
		   state <= next_state;

   always @(*)
	   case(state)
		   4'b0000: next_state = (ready_toggle[idx] == sent_toggle[idx]) ? 4'b0010 : 4'b0001;
			4'b0001: next_state = (!wr_req[pres_route_dir]) ? 4'b0011 : (       // Output direction free
			                      (arb_tab[3*pres_route_dir+:3] != idx) ? 4'b0010 : (   // Output direction occupied by another packet
										  wr_grant[pres_route_dir] ? 4'b1001 : 4'b0111)); // 4'b0101 ));
			4'b0010: next_state = 4'b0000;
			4'b0011: next_state = 4'b0010;
//			4'b0101: next_state = 4'b0111;
			4'b0111: next_state = (timeout[idx*TIMEOUT_W+:TIMEOUT_W] == 0 && pres_route_dir != 0) ? 4'b1111 : 4'b0010;
			4'b1000: next_state = rd_valid ? 4'b1100 : 4'b1000;
			4'b1001: next_state = 4'b1000;
			4'b1010: next_state = 4'b0010;
			4'b1011: next_state = wr_grant[pres_route_dir] ? 4'b1001 : 4'b1010;
			4'b1100: next_state = rd_valid ? 4'b1100 : 4'b1110;
			4'b1110: next_state = 4'b0010;
			4'b1111: next_state = 4'b1011;
			default: next_state = 4'b0010;
		endcase
		
	always @(posedge clk or posedge rst)
		if(rst)
			idx <= 3'b000;
		else if(state == 4'b0010)
			idx <= (idx == 3'b100) ? 3'b000 : idx+3'b001;
			
	always @(posedge clk or posedge rst)
	   if(rst)
			wr_req <= 5'b0_0000;
		else if(state == 4'b0011)
			wr_req[pres_route_dir] <= 1'b1;
		else if(state == 4'b1111 || state == 4'b1001)
			wr_req[pres_route_dir] <= 1'b0;
			
	always @(posedge clk or posedge rst)
		if(rst)
			wr_pack <= 0;
		else if(state == 4'b0011)
			wr_pack[IO_W*pres_route_dir+:IO_W] <= {{(IO_W-NUM_VIA_W){1'b0}}, num_via[NUM_VIA_W*idx+:NUM_VIA_W]};
		else if(state == 4'b1111 || state == 4'b1110)
			wr_pack[IO_W*pres_route_dir+:IO_W] <= 0;
		else if(state == 4'b1100 || state == 4'b1000)
			wr_pack[IO_W*pres_route_dir+:IO_W] <= rd_pack;
			
	always @(posedge clk or posedge rst)
		if(rst)
			wr_valid <= 0;
		else if(state == 4'b1100 || state == 4'b1000)
			wr_valid[pres_route_dir] <= rd_valid;
		else if(state == 4'b1110)
			wr_valid[pres_route_dir] <= 1'b0;
	
	always @(posedge clk or posedge rst)
		if(rst)
			arb_tab <= {15{1'b1}}; // Invalid directions
		else if(state == 4'b0011)
			arb_tab[3*pres_route_dir+:3] <= idx;
		else if(state == 4'b1110 || state == 4'b1010)
			arb_tab[3*pres_route_dir+:3] <= 3'b111;
		
   assign smd_add = idx;
   assign smd_finish = (state == 4'b1110);
//	assign smd_decr = (state == 4'b0101);
	assign smd_reroute = (state == 4'b1010);

   assign rd_fifon = idx;
	assign rd_num_via = num_via[idx*NUM_VIA_W+:NUM_VIA_W];
	assign rd_req = (state == 4'b1001 || state == 4'b1000);

endmodule
