`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date:    14:24:19 05/13/2014 
// Design Name: 
// Module Name:    synth_top 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: 
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
module synth_top_8x8(
 input rst,
 input clk,
 
 input  wr_req_in,
 input  [IO_W-1:0] wr_pack_in,
 input  [2*COORD_SIZE-1:0] wr_add_in,
 output wr_grant_in,
 output wr_reject_in,
 input  wr_valid_in, // INW

 output wr_req_en,
 output [IO_W-1:0] wr_pack_en,
 output  [2*COORD_SIZE-1:0] wr_add_en,
 input  wr_grant_en,
 input  wr_reject_en,
 output wr_valid_en  // ENW
 
    );
parameter WORD_W = 32;    // Width of each word, same as width of the memory interface. Must be even.
parameter MEM_ADD_W = 10;
//parameter PN_Y_COORD = 0;
//parameter PN_X_COORD = 0;
parameter COORD_SIZE = 8;
//parameter FLOP_ON_IENW_REQ = 0;
//parameter FLOP_ON_IENW_GRANT_REJ = 0;
parameter IENW_REQ_FLOP_IVAL = 1;
parameter IENW_GRANT_REJ_FLOP_IVAL = 1;
parameter XSIZE = 8;
parameter YSIZE = 8;

localparam CNWW = IO_W + 2;
localparam line = XSIZE*CNWW;
localparam IO_W = WORD_W/2;

 wire  [3:0] wr_req;
 wire  [4*IO_W-1:0] wr_pack;
 wire [3:0] wr_grant;
 wire  [3:0] wr_valid; // CNW interface - downstream ports, driven by an upstream proc_node




genvar x, y; // ser_no;

wire [CNWW*XSIZE*YSIZE -1:0] cnw_l2r_o, cnw_r2l_o, cnw_u2d_o, cnw_d2u_o; 
wire [CNWW*XSIZE*YSIZE -1:0] cnw_l2r_i, cnw_r2l_i, cnw_u2d_i, cnw_d2u_i; 
wire [XSIZE*YSIZE -1:0] grant_l2r_o, grant_r2l_o, grant_u2d_o, grant_d2u_o;
wire [XSIZE*YSIZE -1:0] grant_l2r_i, grant_r2l_i, grant_u2d_i, grant_d2u_i;

wire [XSIZE*YSIZE :0] ie_grant, ie_reject, ie_req, ie_valid;
wire [(XSIZE*YSIZE+1)*IO_W -1:0] ie_pack;
wire [(XSIZE*YSIZE+1)*2*COORD_SIZE -1: 0] ie_add;

 assign ie_req[0] = wr_req_in;
 assign ie_pack[IO_W-1:0] = wr_pack_in;
 assign ie_add[2*COORD_SIZE-1 :0] = wr_add_in;
 assign wr_grant_in = ie_grant[0];
 assign wr_reject_in = ie_reject[0];
 assign ie_valid[0] = wr_valid_in; // INW

 assign wr_req_en = ie_req[XSIZE*YSIZE];
 assign wr_pack_en = ie_pack[(XSIZE*YSIZE+1)*IO_W -1: XSIZE*YSIZE*IO_W]; // [IO_W-1:0]
 assign wr_add_en = ie_add[(XSIZE*YSIZE+1)*2*COORD_SIZE -1: XSIZE*YSIZE*2*COORD_SIZE];
 assign ie_grant[XSIZE*YSIZE] = wr_grant_en;
 assign ie_reject[XSIZE*YSIZE] = wr_reject_en;
 assign wr_valid_en = ie_valid[XSIZE*YSIZE]; 

generate
for(y = 0; y < YSIZE; y = y+1)
begin
   for(x = 0; x < XSIZE; x = x+1)
	begin
	   localparam ser = x + y*XSIZE;
		localparam cnwr = ser*CNWW;
		localparam ieser = (y%2) ? y*XSIZE + (XSIZE-1-x): (x+y*XSIZE);
		
		proc_node_with_mcu #(
			.PN_Y_COORD(y),
			.PN_X_COORD(x),
			.COORD_SIZE(COORD_SIZE),
			.FLOP_ON_IENW_REQ(1),
			.FLOP_ON_IENW_GRANT_REJ((ieser%6 == 0) || (y==7 && x==0))
		) proc_node0(
		.rst(rst),
		.clk(clk),

		.ie_grant_ds(ie_grant[ieser]),
		.ie_reject_ds(ie_reject[ieser]),
		.ie_req_ds(ie_req[ieser]),
		.ie_pack_ds(ie_pack[(ieser+1)*IO_W -1: ieser*IO_W]),
		.ie_add_ds(ie_add[(ieser+1)*(2*COORD_SIZE) -1 : ieser*2*COORD_SIZE]),
		.ie_valid_ds(ie_valid[ieser]), // IENW interface - downstream ports, driven by an upstream proc_node
 
		.ie_grant_us(ie_grant[ieser+1]),
		.ie_reject_us(ie_reject[ieser+1]),
		.ie_req_us(ie_req[ieser+1]),
		.ie_pack_us(ie_pack[(ieser+2)*IO_W -1: (ieser+1)*IO_W]),
		.ie_add_us(ie_add[(ieser+2)*(2*COORD_SIZE) -1 : (ieser+1)*2*COORD_SIZE]),
		.ie_valid_us(ie_valid[ieser+1]), // IENW interface - upstream ports, to drive a downstream proc_node
 
		.wr_grant_ds({grant_u2d_o[ser], grant_l2r_o[ser], grant_d2u_o[ser], grant_r2l_o[ser]}),
		.wr_pack_ds({cnw_d2u_i[cnwr+CNWW-1:cnwr+2], cnw_r2l_i[cnwr+CNWW-1:cnwr+2], cnw_u2d_i[cnwr+CNWW-1:cnwr+2], cnw_l2r_i[cnwr+CNWW-1:cnwr+2]}),
		.wr_req_ds({cnw_d2u_i[cnwr+1], cnw_r2l_i[cnwr+1], cnw_u2d_i[cnwr+1], cnw_l2r_i[cnwr+1]}),
		.wr_valid_ds({cnw_d2u_i[cnwr], cnw_r2l_i[cnwr], cnw_u2d_i[cnwr], cnw_l2r_i[cnwr]}), // CNW interface - downstream ports, driven by an upstream proc_node
 
		.wr_grant_us({grant_d2u_i[ser], grant_r2l_i[ser], grant_u2d_i[ser], grant_l2r_i[ser]}),
		.wr_pack_us({cnw_u2d_o[cnwr+CNWW-1:cnwr+2], cnw_l2r_o[cnwr+CNWW-1:cnwr+2], cnw_d2u_o[cnwr+CNWW-1:cnwr+2], cnw_r2l_o[cnwr+CNWW-1:cnwr+2]}),
		.wr_req_us({cnw_u2d_o[cnwr+1], cnw_l2r_o[cnwr+1], cnw_d2u_o[cnwr+1], cnw_r2l_o[cnwr+1]}),
		.wr_valid_us({cnw_u2d_o[cnwr], cnw_l2r_o[cnwr], cnw_d2u_o[cnwr], cnw_r2l_o[cnwr]})  // CNW interface - upstream ports, to drive a downstream proc_node
);

      if(x == 0)
		begin
		   assign grant_l2r_i[ser] = grant_r2l_o[ser];
			assign cnw_l2r_i[cnwr+CNWW-1:cnwr] = cnw_r2l_o[cnwr+CNWW-1:cnwr];
		end
		else
		begin
		   assign grant_l2r_i[ser] = grant_l2r_o[ser-1];
			assign cnw_l2r_i[cnwr+CNWW-1:cnwr] = cnw_l2r_o[cnwr-1:cnwr-CNWW];
		end
			
		if(x == XSIZE-1)
		begin
		   assign grant_r2l_i[ser] = grant_l2r_o[ser];
			assign cnw_r2l_i[cnwr+CNWW-1:cnwr] = cnw_l2r_o[cnwr+CNWW-1:cnwr];
		end
		else
		begin
			assign grant_r2l_i[ser] = grant_r2l_o[ser+1];
			assign cnw_r2l_i[cnwr+CNWW-1:cnwr] = cnw_r2l_o[cnwr+2*CNWW-1:cnwr+CNWW];
		end
			
		if(y == 0)
		begin
		   assign grant_d2u_i[ser] = grant_u2d_o[ser];
			assign cnw_d2u_i[cnwr+CNWW-1:cnwr] = cnw_u2d_o[cnwr+CNWW-1:cnwr];
		end
		else
		begin
			assign grant_d2u_i[ser] = grant_d2u_o[ser-XSIZE];
			assign cnw_d2u_i[cnwr+CNWW-1:cnwr] = cnw_d2u_o[cnwr+CNWW-line-1 : cnwr-line];
		end
			
		if(y == YSIZE-1)
		begin
		   assign grant_u2d_i[ser] = grant_d2u_o[ser];
			assign cnw_u2d_i[cnwr+CNWW-1:cnwr] = cnw_d2u_o[cnwr+CNWW-1:cnwr];
		end
		else
		begin
			assign grant_u2d_i[ser] = grant_u2d_i[ser+XSIZE];
			assign cnw_u2d_i[cnwr+CNWW-1:cnwr] = cnw_u2d_o[cnwr+CNWW+line-1:cnwr+line];
		end
   end
end
endgenerate

endmodule
