`timescale 1ns / 1ps
/*
Description:

1. Writes 2 packets into FIFO #4
2. Reads both packets out of FIFO #4, but with reduced number of via points
3. Reading out of first packet and writing in of second packet overlap
4. Send-module not used. Only FIFO-controller and memory used.

Things verified:
1. Data integrity of the packets
- including the replacement of the old number of via by new number of via
- required number of via are read out

2. Updation of receive-list
- Initialization by read FSM
- Updation by both read and write FSM's

*/

module test_bench2(

    );

   parameter WORD_W = 32;    // Width of each word, same as width of the memory interface. Must be even.
   parameter MEM_ADD_W = 10;
   parameter HEADER_SIZE = 5; // size of ftr packet header in number of words
	parameter FIFO_DEPTH_W = 7; // Depth of FIFO in number of words = (2^FIFO_DEPTH_W -1) : -1 for distinguishing empty/full
   parameter NUM_VIA_W = FIFO_DEPTH_W; // Max number of via points in a packet = 2^NUM_VIA_W

   localparam IO_W = WORD_W/2;

   reg clk;
	reg rst;

 reg [4:0] wr_req;
 reg [5*IO_W-1:0] wr_pack;
 wire [4:0] wr_grant;
 reg [4:0] wr_valid; // IENW & CNW interface

 reg rd_req;
 wire [IO_W-1:0] rd_pack;
 wire   rd_valid;
 reg [2:0] rd_fifon;
 reg [NUM_VIA_W-1:0] rd_num_via; // send module interface

 wire [WORD_W-1:0] mem_od;
 wire [WORD_W-1:0] mem_id;
 wire              mem_csel;
 wire [MEM_ADD_W-1:0] mem_add;
 wire      mem_wr;

   initial
	begin
	   clk <= 1'b0;
      forever
	      #5 clk <= !clk;
   end
	
	initial
	begin
	   rst <= 1'b1;
		rd_req <= 1'b0;
		wr_req <= 0;
		wr_pack <= 0;
		wr_valid <= 0;
		rd_fifon <= 0;
		rd_num_via <= 0;
//		mem_od <= 0;
		
		#20 rst <= 1'b0;

// Write a packet with 10 via points
		#20 wr_pack[79:64] <= 10;
		    wr_req[4] <= 1;
			 
		@(posedge wr_grant[4]);
		@(posedge clk);
		wr_req = 0;
		
		@(posedge clk);
		@(posedge clk);
		wr_pack[79:64] <= 16'h550a;
		wr_valid[4] <= 1;
		
		#300 wr_valid <= 0;

// Read out the same packet with 8 via points
		#35 		rd_num_via <= 8;
		rd_fifon <= 4;
		@(posedge clk);
		rd_req <= 1'b1;


 	   @(posedge rd_valid);
		@(posedge clk);
		rd_req <= 0;

// In parallel, write a packet with 4 via points into the same FIFO
		wr_pack[79:64] <= 4;
		    wr_req[4] <= 1;

		@(posedge wr_grant[4]);
		@(posedge clk);
		wr_req = 0;
		
		@(posedge clk);
		@(posedge clk);
		wr_pack[79:64] <= 16'haa04;
		wr_valid[4] <= 1;

		#180 wr_valid <= 0;

// Read out the second packet - 2 vias only
		#35 		rd_num_via <= 2;
		rd_fifon <= 4;
		@(posedge clk);
		rd_req <= 1'b1;

		@(posedge rd_valid);
		@(posedge clk);
		rd_req <= 0;
					 
		#400 $finish;
		
	end
	
	
fifo_ctrl fifo_ctrl_inst (
	.rst(rst),
	.clk(clk),

 .wr_req(wr_req),
 .wr_pack(wr_pack),
 .wr_grant(wr_grant),
 .wr_valid(wr_valid), // IENW & CNW interface

 .rd_req(rd_req),
 .rd_pack(rd_pack),
 .rd_valid(rd_valid),
 .rd_fifon(rd_fifon),
 .rd_num_via(rd_num_via), // send module interface

 .mem_od(mem_od),
 .mem_id(mem_id),
 .mem_csel(mem_csel),
 .mem_add(mem_add),
 .mem_wr(mem_wr)
	);

block_ram block_ram_inst (
 .clockA(clk),
 .clockB(1'b0),
 
 .odataA(mem_od),
 .addA(mem_add),
 .idataA(mem_id),
 .enableA(mem_csel),
 .write_enableA(mem_wr),

 .odataB(),
 .addB(),
 .idataB(),
 .enableB(),
 .write_enableB()
);

endmodule
