`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date:    20:33:04 05/23/2014 
// Design Name: 
// Module Name:    test_bench4 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: 
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
module test_bench4(
    );
parameter TIMEOUT_W = 10;
parameter WORD_W = 32;    // Width of each word, same as width of the memory interface. Must be even.
parameter MEM_ADD_W = 10;
parameter HEADER_SIZE = 5; // size of ftr packet header in number of words
parameter FIFO_DEPTH_W = 7; // Depth of FIFO in number of words = (2^FIFO_DEPTH_W -1) : -1 for distinguishing empty/full
parameter COORD_SIZE = 8;
parameter NUM_VIA_W = FIFO_DEPTH_W;
parameter REC_LIST_ADD = 5 * (2 ** FIFO_DEPTH_W);
parameter FLOP_ON_IENW_REQ = 0;
parameter FLOP_ON_IENW_GRANT_REJ = 0;
parameter TIMEOUT_COUNT_PREC = 0; // Timeout is counted every 2**TIMEOUT_COUNT_PREC clocks

localparam SEND_LIST_BASE = 2**MEM_ADD_W;
localparam IO_W = WORD_W/2;


reg rst, clk;

reg  ie_req_ds;
reg  [IO_W-1:0] ie_pack_ds;
reg  [(COORD_SIZE*2)-1:0] ie_add_ds;
wire ie_grant_ds;
wire ie_reject_ds;
reg  ie_valid_ds; // IENW interface - downstream ports, driven by an upstream proc_node

wire  ie_req_us;
wire  [IO_W-1:0] ie_pack_us;
wire  [(COORD_SIZE*2)-1:0] ie_add_us;
reg ie_grant_us;
reg ie_reject_us;
wire  ie_valid_us; // IENW interface - upstream ports, to drive a downstream proc_node

wire [3:0] wr_req_ds2;
wire  [4*IO_W-1:0] wr_pack_ds2;
wire [3:0] wr_grant_ds2;
wire  [3:0] wr_valid_ds2;

wire [3:0] wr_req_us2;
wire  [4*IO_W-1:0] wr_pack_us2;
reg [3:0] wr_grant_us2;
wire  [3:0] wr_valid_us2;

 reg  [MEM_ADD_W:0] cpu_add2;
 reg   cpu_wr2;
 reg   cpu_cs2;
 wire [WORD_W-1:0] cpu_data_i2;
 wire  [WORD_W-1:0] cpu_data_o2;

 reg ready_toggle2;
 reg [TIMEOUT_W-1:0] timeout2;
 reg [2:0] route_dir2;
 reg [NUM_VIA_W-1:0] num_via2;

initial
begin
   clk <= 1'b0;
	
	forever
	   #5 clk <= ~clk;
end
 
 assign cpu_data_o2 = {ready_toggle2, timeout2, route_dir2, num_via2};

reg flag;
// Packet source
initial
begin
   rst <= 1'b1;
   ie_req_ds <= 1'b0;
	ie_pack_ds <= 10;
	ie_add_ds <= 16'h0002;
	ie_valid_ds <= 1'b0;
	flag <= 0;
	
	#20 rst <= 1'b0;
	
	#40 ie_pack_ds <= 10;
	   ie_req_ds <= 1'b1;
		
	@(posedge ie_grant_ds);
	@(posedge clk)
	   ie_req_ds <= 1'b0;
		ie_valid_ds <= 1'b1;
		ie_pack_ds <= 16'h0f_0a;
		
	#300 ie_valid_ds <= 0;

// Filler Packet 1
	@(posedge clk);
	@(posedge clk)
	   ie_req_ds <= 1'b1;
		ie_pack_ds  <= 6;
		
	@(posedge ie_grant_ds);
	@(posedge clk)
	   ie_req_ds <= 1'b0;
		ie_valid_ds <= 1'b1;
		ie_pack_ds <= 16'h1f_06;
		
	#220 ie_valid_ds <= 0;
	
// Filler Packet 2
	@(posedge clk);
	@(posedge clk)
	   ie_req_ds <= 1'b1;
		ie_pack_ds  <= 6;
		
	@(posedge ie_grant_ds);
	@(posedge clk)
	   ie_req_ds <= 1'b0;
		ie_valid_ds <= 1'b1;
		ie_pack_ds <= 16'h2f_06;
		
	#220 ie_valid_ds <= 0;
	
// Filler Packet 3
	@(posedge clk);
	@(posedge clk)
	   ie_req_ds <= 1'b1;
		ie_pack_ds  <= 6;
		
	@(posedge ie_grant_ds);
	@(posedge clk)
	   ie_req_ds <= 1'b0;
		ie_valid_ds <= 1'b1;
		ie_pack_ds <= 16'h3f_06;
		
	#220 ie_valid_ds <= 0;
	
// Filler Packet 4
	@(posedge clk);
	@(posedge clk)
	   ie_req_ds <= 1'b1;
		ie_pack_ds  <= 6;
		
	@(posedge ie_grant_ds);
	@(posedge clk)
	   ie_req_ds <= 1'b0;
		ie_valid_ds <= 1'b1;
		ie_pack_ds <= 16'h4f_06;
		
	#220 ie_valid_ds <= 0;
	
// Filler Packet 4
	@(posedge clk);
	@(posedge clk)
	   ie_req_ds <= 1'b1;
		ie_pack_ds  <= 6;
		
	@(posedge ie_grant_ds);
	@(posedge clk)
	   ie_req_ds <= 1'b0;
		ie_valid_ds <= 1'b1;
		ie_pack_ds <= 16'h5f_06;
		
	#220 ie_valid_ds <= 0;
	
// Filler Packet 6
	@(posedge clk);
	@(posedge clk)
	   ie_req_ds <= 1'b1;
		ie_pack_ds  <= 6;
		
	@(posedge ie_grant_ds);
	@(posedge clk)
	   ie_req_ds <= 1'b0;
		ie_valid_ds <= 1'b1;
		ie_pack_ds <= 16'h6f_06;
		
	#220 ie_valid_ds <= 0;
	
// Filler Packet 7
	@(posedge clk);
	@(posedge clk)
	   ie_req_ds <= 1'b1;
		ie_pack_ds  <= 6;
		
	@(posedge ie_grant_ds);
	@(posedge clk)
	   ie_req_ds <= 1'b0;
		ie_valid_ds <= 1'b1;
		ie_pack_ds <= 16'h7f_06;
		
	#220 ie_valid_ds <= 0;
	
// Filler Packet 8
	@(posedge clk);
	@(posedge clk)
	   ie_req_ds <= 1'b1;
		ie_pack_ds  <= 6;
		
	@(posedge ie_grant_ds);
	@(posedge clk)
	   ie_req_ds <= 1'b0;
		ie_valid_ds <= 1'b1;
		ie_pack_ds <= 16'h8f_06;
		
	#220 ie_valid_ds <= 0;
	
// Filler Packet 9
	@(posedge clk);
	@(posedge clk)
	   ie_req_ds <= 1'b1;
		ie_pack_ds  <= 6;
		
	@(posedge ie_grant_ds);
	@(posedge clk)
	   ie_req_ds <= 1'b0;
		ie_valid_ds <= 1'b1;
		ie_pack_ds <= 16'h9f_06;
		
	#220 ie_valid_ds <= 0;
	
// Filler Packet 10
	@(posedge clk);
	@(posedge clk)
	   ie_req_ds <= 1'b1;
		ie_pack_ds  <= 6;
		
	@(posedge ie_grant_ds);
	@(posedge clk)
	   ie_req_ds <= 1'b0;
		ie_valid_ds <= 1'b1;
		ie_pack_ds <= 16'haf_06;
		
	#220 ie_valid_ds <= 0;
	
// Filler Packet 11
	@(posedge clk);
	@(posedge clk)
	   ie_req_ds <= 1'b1;
		ie_pack_ds  <= 6;
		
	@(posedge ie_grant_ds);
	@(posedge clk)
	   ie_req_ds <= 1'b0;
		ie_valid_ds <= 1'b1;
		ie_pack_ds <= 16'hbf_06;
		
	#220 ie_valid_ds <= 0;
	
// Filler Packet 12
	@(posedge clk);
	@(posedge clk)
	   ie_req_ds <= 1'b1;
		ie_pack_ds  <= 6;

/*	Check for reject */
   @(posedge ie_reject_ds)
	   flag <= 1;
	

end

// CPU monitor
initial
begin
 cpu_add2 <= 11'b010_1000_0000; // FIFO_0 entry in receive list
 cpu_wr2 <= 0;
 cpu_cs2 <= 1;
// reg  [WORD_W-1:0] cpu_data_o2;

 @(cpu_data_i2);
 @(cpu_data_i2);
 @(posedge clk)
    cpu_cs2 <= 0;

 @(posedge clk)
    cpu_add2 <= 11'b100_0000_0000; // Send-list entry for FIFO_0
	 cpu_wr2 <= 1'b1;
	 cpu_cs2 <= 1'b1;
	 ready_toggle2 <= 1'b1;
	 timeout2 <= 80;
	 route_dir2 <= 1;
	 num_via2 <= 3;
	 
  @(posedge clk)
     cpu_wr2 <= 1'b0;
     cpu_add2 <= 11'b010_1000_0000; // FIFO_0 entry in receive list
	  
  @(posedge flag) // Received a reject from FIFO => FIFO full
    cpu_add2 <= 11'b100_0000_0000; // Send-list entry for FIFO_0
	 cpu_wr2 <= 1'b1;
	 cpu_cs2 <= 1'b1;
	 ready_toggle2 <= 1'b0;
	 timeout2 <= 10;
	 route_dir2 <= 1;
	 num_via2 <= 5;

  @(posedge clk)
     cpu_wr2 <= 1'b0;
end

// CNW monitor
initial
begin
   wr_grant_us2 <= 0;
	
	@(posedge wr_req_us2[0]);
	@(posedge clk)
	   wr_grant_us2[0] <= 1;
		
	@(posedge wr_valid_us2[0]);
	@(posedge clk)
	   wr_grant_us2[0] <= 0;
		
	@(posedge wr_req_us2[0]); // The request at the very end of simulation
	@(negedge wr_req_us2[0])  // Just timed-out
	   wr_grant_us2[0] <= 1;
		
	@(posedge wr_valid_us2[0]);
	@(posedge clk)
	   wr_grant_us2[0] <= 0;

	#100 $finish;

end

proc_node_with_ie #(
.PN_Y_COORD(0),
.PN_X_COORD(2)
) proc_node_inst(
 .rst(rst),
 .clk(clk),

 .ie_req_ds(ie_req_ds),
 .ie_pack_ds(ie_pack_ds),
 .ie_add_ds(ie_add_ds),
 .ie_grant_ds(ie_grant_ds),
 .ie_reject_ds(ie_reject_ds),
 .ie_valid_ds(ie_valid_ds), // IENW interface - downstream ports, driven by an upstream proc_node
 
 .ie_req_us(ie_req_us),
 .ie_pack_us(ie_pack_us),
 .ie_add_us(ie_add_us),
 .ie_grant_us(0),
 .ie_reject_us(0),
 .ie_valid_us(ie_valid_us), // IENW interface - upstream ports, to drive a downstream proc_node

 .wr_req_ds(0),
 .wr_pack_ds(0),
 .wr_grant_ds(wr_grant_ds2),
 .wr_valid_ds(0), // CNW interface - downstream ports, driven by an upstream proc_node
 
 .wr_req_us(wr_req_us2),
 .wr_pack_us(wr_pack_us2),
 .wr_grant_us(wr_grant_us2),
 .wr_valid_us(wr_valid_us2), // CNW interface - upstream ports, to drive a downstream proc_node
 
 .cpu_add(cpu_add2),
 .cpu_wr(cpu_wr2),
 .cpu_cs(cpu_cs2),
 .cpu_data_i(cpu_data_i2),
 .cpu_data_o(cpu_data_o2) // Processor interface
);

endmodule
