`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date:    18:08:42 05/23/2014 
// Design Name: 
// Module Name:    test_bench7 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: 
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
module test_bench7(
    );
parameter TIMEOUT_W = 10;
parameter WORD_W = 32;    // Width of each word, same as width of the memory interface. Must be even.
parameter MEM_ADD_W = 10;
parameter HEADER_SIZE = 5; // size of ftr packet header in number of words
parameter FIFO_DEPTH_W = 7; // Depth of FIFO in number of words = (2^FIFO_DEPTH_W -1) : -1 for distinguishing empty/full
parameter COORD_SIZE = 8;
parameter NUM_VIA_W = FIFO_DEPTH_W;
parameter REC_LIST_ADD = 5 * (2 ** FIFO_DEPTH_W);
parameter FLOP_ON_IENW_REQ = 0;
parameter FLOP_ON_IENW_GRANT_REJ = 0;
parameter TIMEOUT_COUNT_PREC = 0; // Timeout is counted every 2**TIMEOUT_COUNT_PREC clocks

localparam SEND_LIST_BASE = 2**MEM_ADD_W;
localparam IO_W = WORD_W/2;

reg rst, clk;

reg  ie_req_ds;
reg  [IO_W-1:0] ie_pack_ds;
reg  [(COORD_SIZE*2)-1:0] ie_add_ds;
wire ie_grant_ds;
wire ie_reject_ds;
reg  ie_valid_ds; // IENW interface - downstream ports, driven by an upstream proc_node

wire  ie_req_int;
wire  [IO_W-1:0] ie_pack_int;
wire  [(COORD_SIZE*2)-1:0] ie_add_int;
wire ie_grant_int;
wire ie_reject_int;
wire  ie_valid_int; // IENW interface - internal

wire  ie_req_us;
wire  [IO_W-1:0] ie_pack_us;
wire  [(COORD_SIZE*2)-1:0] ie_add_us;
reg ie_grant_us;
reg ie_reject_us;
wire  ie_valid_us; // IENW interface - upstream ports, to drive a downstream proc_node


wire [3:0] wr_req_ds1;
wire  [4*IO_W-1:0] wr_pack_ds1;
wire [3:0] wr_grant_ds1;
wire  [3:0] wr_valid_ds1;

wire [3:0] wr_req_us1;
wire  [4*IO_W-1:0] wr_pack_us1;
wire [3:0] wr_grant_us1;
wire  [3:0] wr_valid_us1;

wire [3:0] wr_req_ds2;
wire  [4*IO_W-1:0] wr_pack_ds2;
wire [3:0] wr_grant_ds2;
wire  [3:0] wr_valid_ds2;

wire [3:0] wr_req_us2;
wire  [4*IO_W-1:0] wr_pack_us2;
wire [3:0] wr_grant_us2;
wire  [3:0] wr_valid_us2;

 reg  [MEM_ADD_W:0] cpu_add1;
 reg   cpu_wr1;
 reg   cpu_cs1;
 wire [WORD_W-1:0] cpu_data_i1;
 wire  [WORD_W-1:0] cpu_data_o1;

 reg  [MEM_ADD_W:0] cpu_add2;
 reg   cpu_wr2;
 reg   cpu_cs2;
 wire [WORD_W-1:0] cpu_data_i2;
 wire  [WORD_W-1:0] cpu_data_o2;

assign wr_req_ds2[3:1] = wr_req_us2[3:1];
assign wr_pack_ds2[4*IO_W -1: IO_W] = wr_pack_us2[4*IO_W -1: IO_W];
assign wr_valid_ds2[3:1] = wr_valid_us2[3:1];
assign wr_grant_us2[3:1] = wr_grant_ds2[3:1];

assign wr_req_ds1[1:0] = wr_req_us1[1:0];
assign wr_pack_ds1[2*IO_W -1: 0] = wr_pack_us1[2*IO_W -1:0];
assign wr_valid_ds1[1:0] = wr_valid_us1[1:0];
assign wr_grant_us1[1:0] = wr_grant_ds1[1:0];
assign wr_req_ds1[3] = wr_req_us1[3];
assign wr_pack_ds1[4*IO_W -1: 3*IO_W] = wr_pack_us1[4*IO_W -1: 3*IO_W];
assign wr_valid_ds1[3] = wr_valid_us1[3];
assign wr_grant_us1[3] = wr_grant_ds1[3];

assign wr_req_ds2[0] = wr_req_us1[2];
assign wr_pack_ds2[IO_W-1 :0] = wr_pack_us1[3*IO_W -1: 2*IO_W];
assign wr_valid_ds2[0] = wr_valid_us1[2];
assign wr_grant_us1[2] = wr_grant_ds2[0];

assign wr_req_ds1[2] = wr_req_us2[0];
assign wr_pack_ds1[3*IO_W-1 :2*IO_W] = wr_pack_us2[IO_W -1: 0];
assign wr_valid_ds1[2] = wr_valid_us2[0];
assign wr_grant_us2[0] = wr_grant_ds1[2];

initial
begin
   clk <= 1'b0;
	
	forever
	   #5 clk <= ~clk;
end

// Packet source
initial
begin
   rst <= 1'b1;
   ie_req_ds <= 1'b0;
	ie_pack_ds <= 3;
	ie_add_ds <= 16'h0002;
	ie_valid_ds <= 1'b0;
	
	#20 rst <= 1'b0;
	
	#100 ie_req_ds <= 1'b1;
	
	@(posedge ie_grant_ds);
	@(posedge clk)
	   ie_req_ds <= 1'b0;
		
	#100;
	@(posedge clk)
	   ie_valid_ds <= 1'b1;
		ie_pack_ds <= 16'hff_03;
		
	#160 ie_valid_ds <= 0;
end

// Node2: FIFO monitor
 reg ready_toggle2;
 reg [TIMEOUT_W-1:0] timeout2;
 reg [2:0] route_dir2;
 reg [NUM_VIA_W-1:0] num_via2;
 
 assign cpu_data_o2 = {ready_toggle2, timeout2, route_dir2, num_via2};

initial
begin
 cpu_add2 <= 11'b010_1000_0000; // FIFO_0 entry in receive list
 cpu_wr2 <= 0;
 cpu_cs2 <= 1;
// reg  [WORD_W-1:0] cpu_data_o2;

 @(cpu_data_i2);
 @(cpu_data_i2);
 @(posedge clk)
    cpu_cs2 <= 0;

 #100;
 @(posedge clk)
    cpu_add2 <= 11'b100_0000_0000; // Send-list entry for FIFO_0
	 cpu_wr2 <= 1'b1;
	 cpu_cs2 <= 1'b1;
	 ready_toggle2 <= 1'b1;
	 timeout2 <= 80;
	 route_dir2 <= 1;
	 num_via2 <= 3;
	 
  @(posedge clk)
     cpu_wr2 <= 1'b0;

end


// Node1: FIFO monitor
 reg ready_toggle1;
 reg [TIMEOUT_W-1:0] timeout1;
 reg [2:0] route_dir1;
 reg [NUM_VIA_W-1:0] num_via1;
 
 assign cpu_data_o1 = {ready_toggle1, timeout1, route_dir1, num_via1};

initial
begin
 cpu_add1 <= 11'b010_1000_0011; // FIFO_3 entry in receive list
 cpu_wr1 <= 0;
 cpu_cs1 <= 1;
// reg  [WORD_W-1:0] cpu_data_o2;

 @(cpu_data_i1);
 @(cpu_data_i1);
 @(posedge clk)
    cpu_cs1 <= 0;

 #110;
 @(posedge clk)
    cpu_add1 <= 11'b100_0000_0011; // Send-list entry for FIFO_3
	 cpu_wr1 <= 1'b1;
	 cpu_cs1 <= 1'b1;
	 ready_toggle1 <= 1'b1;
	 timeout1 <= 80;
	 route_dir1 <= 0;
	 num_via1 <= 1;
	 
  @(posedge clk)
     cpu_wr1 <= 1'b0;

end

// Packet sink
initial
begin
   ie_grant_us <= 0;
   ie_reject_us <= 0;
	
	@(posedge ie_req_us);
	#50;
	@(posedge clk)
	   ie_reject_us <= 1;
	@(posedge clk)
	   ie_reject_us <= 0;
		
	@(posedge ie_req_us);
	#60;
	@(posedge clk)
	   ie_grant_us <= 1;
		
	@(posedge ie_valid_us);
	@(posedge clk)
	   ie_grant_us <= 0;
		
  #200 $finish;
	
end

proc_node_with_ie # (
.PN_Y_COORD(0),
.PN_X_COORD(2)
) proc_node_2(
 .rst(rst),
 .clk(clk),

 .ie_req_ds(ie_req_int),
 .ie_pack_ds(ie_pack_int),
 .ie_add_ds(ie_add_int),
 .ie_grant_ds(ie_grant_int),
 .ie_reject_ds(ie_reject_int),
 .ie_valid_ds(ie_valid_int), // IENW interface - downstream ports, driven by an upstream proc_node
 
 .ie_req_us(ie_req_us),
 .ie_pack_us(ie_pack_us),
 .ie_add_us(ie_add_us),
 .ie_grant_us(ie_grant_us),
 .ie_reject_us(ie_reject_us),
 .ie_valid_us(ie_valid_us), // IENW interface - upstream ports, to drive a downstream proc_node

 .wr_req_ds(wr_req_ds2),
 .wr_pack_ds(wr_pack_ds2),
 .wr_grant_ds(wr_grant_ds2),
 .wr_valid_ds(wr_valid_ds2), // CNW interface - downstream ports, driven by an upstream proc_node
 
 .wr_req_us(wr_req_us2),
 .wr_pack_us(wr_pack_us2),
 .wr_grant_us(wr_grant_us2),
 .wr_valid_us(wr_valid_us2), // CNW interface - upstream ports, to drive a downstream proc_node
 
 .cpu_add(cpu_add2),
 .cpu_wr(cpu_wr2),
 .cpu_cs(cpu_cs2),
 .cpu_data_i(cpu_data_i2),
 .cpu_data_o(cpu_data_o2) // Processor interface
);

proc_node_with_ie # (
.PN_Y_COORD(0),
.PN_X_COORD(1)
) proc_node_1(
 .rst(rst),
 .clk(clk),

 .ie_req_ds(ie_req_ds),
 .ie_pack_ds(ie_pack_ds),
 .ie_add_ds(ie_add_ds),
 .ie_grant_ds(ie_grant_ds),
 .ie_reject_ds(ie_reject_ds),
 .ie_valid_ds(ie_valid_ds), // IENW interface - downstream ports, driven by an upstream proc_node
 
 .ie_req_us(ie_req_int),
 .ie_pack_us(ie_pack_int),
 .ie_add_us(ie_add_int),
 .ie_grant_us(ie_grant_int),
 .ie_reject_us(ie_reject_int),
 .ie_valid_us(ie_valid_int), // IENW interface - upstream ports, to drive a downstream proc_node

 .wr_req_ds(wr_req_ds1),
 .wr_pack_ds(wr_pack_ds1),
 .wr_grant_ds(wr_grant_ds1),
 .wr_valid_ds(wr_valid_ds1), // CNW interface - downstream ports, driven by an upstream proc_node
 
 .wr_req_us(wr_req_us1),
 .wr_pack_us(wr_pack_us1),
 .wr_grant_us(wr_grant_us1),
 .wr_valid_us(wr_valid_us1), // CNW interface - upstream ports, to drive a downstream proc_node
 
 .cpu_add(cpu_add1),
 .cpu_wr(cpu_wr1),
 .cpu_cs(cpu_cs1),
 .cpu_data_i(cpu_data_i1),
 .cpu_data_o(cpu_data_o1) // Processor interface
);

endmodule
