#include "fifo.h"
#include "simparam.h"

fifo::fifo(sc_module_name name, int size_)
          : sc_channel(name), size(size_)
{
  data = new ftr_t* [size];

  num_write = byte_write = 0;
  num_read  = byte_read  = 0;
}

fifo::~fifo() {

  delete[] data;

}


void fifo::write (ftr_t* c) {
  if (num_write-num_read == size)  wait(read_ev);

  data[num_write % size] = c;

  num_write++;
  byte_write += BYTES_PER_VIA*(c->no_via) + BYTES_OHEAD;

//  if(c->debug) cout << byte_write-byte_read << endl;

  write_ev.notify();
}


ftr_t* fifo::read () {
  ftr_t* c;

  if (num_read == num_write)  wait(write_ev);

  c = data[num_read % size];

  num_read++;
  byte_read += BYTES_PER_VIA*(c->no_via) + BYTES_OHEAD;

//  if(c->debug) cout << byte_write-byte_read << endl;

  read_ev.notify();

  return c;
}

void fifo::read_finalize(int old_bytes) {
   num_read++;
   byte_read += old_bytes;
   read_ev.notify();
}


ftr_t* fifo::read_init () {
  ftr_t* c;

  if (num_read == num_write)  wait(write_ev);

  c = data[num_read % size];

//  num_read++;
//  byte_read += BYTES_PER_VIA*(c->no_via) + BYTES_OHEAD;

//  read_ev.notify();

  return c;
}

// reset buffer counters
void fifo::reset() { 
  num_read = num_write = 0;
}

// buffer fill
int fifo::num_available() {
  return num_write-num_read;
}


int fifo::bytes_free() {
   return BYTES_FIFO_DEPTH-(byte_write-byte_read);
}
