#include <systemc.h>
#include "simparam.h"
#include "network.h"
#include "global.h"
#include <math.h>

float network::get_network_load() {
   ofstream logfile;
   sc_time  time_stamp = sc_time_stamp();
   int n;
   sc_time busy_time(0, SC_NS);
   const sc_time zero(0, SC_NS);


   logfile.open("network_load.log");
   logfile << "Total time elapsed = " << time_stamp << endl;


   n = 0;
   for(int y=0; y< YSIZE; y++) {
      for(int x=0; x< XSIZE; x++) {
         logfile << x << "," << y << ":" << pn[n]-> total_busy_time << "\t";
         busy_time += pn[n]-> total_busy_time;
         n++;
      }
   logfile << endl;
   }

   busy_time = busy_time/n;

   logfile << endl << "Average busy time of each node = " << busy_time << endl;
   logfile << "Average loading = " << busy_time/time_stamp*100 << "%" << endl;


   n = 0;
   sc_time busy_time_spread(0, SC_NS); // sum of absolute difference from mean / mean
   sc_time diff;

   for(int y=0; y< YSIZE; y++)
      for(int x=0; x< XSIZE; x++, n++) {
         if(pn[n]-> total_busy_time > busy_time)
            diff = (pn[n]-> total_busy_time - busy_time);
         else
            diff = busy_time - pn[n]-> total_busy_time;

         busy_time_spread += diff;
      }

   busy_time_spread /= n;

   logfile << "Average spread of busy time around the mean = " << busy_time_spread << " = " << 100*busy_time_spread/busy_time << "% of average" << endl;


   logfile << "Processing grid dimension: " << XSIZE << " x " << YSIZE << endl;
   logfile << "Folding level: " << XLEVELS << " x " << YLEVELS << endl;

   logfile.close();
   return (busy_time/time_stamp);
}


   float network::gendata(float x, float y) {
      const float K1=1.15;
      const float x1=5;
      const float y1=16;

      const float K2=1.05;
      const float x2=13;
      const float y2=14;

      const float K3=1.1;
      const float x3=7;
      const float y3=8;

      float f1 = 1.0 / powf(K1, (powf(powf(x-x1,2) + powf(y-y1,2),K1)));
      float f2 = 1.0 / powf(K2, (powf(powf(x-x2,2) + powf(y-y2,2),K2)));
      float f3 = 1.0 / powf(K3, (powf(powf(x-x3,2) + powf(y-y3,2),K3)));

      return f1+0.7*f2+0.6*f3;
   }


network::network (sc_module_name name) : sc_module(name) {

   rt  = new con_t* [XSIZE * YSIZE];
   lt  = new con_t* [XSIZE * YSIZE];
   up  = new con_t* [XSIZE * YSIZE];
   dn  = new con_t* [XSIZE * YSIZE];
   enw = new con_t* [XSIZE * YSIZE];
   inw = new con_t* [XSIZE * YSIZE];
   econt = new enw_cont("ENW_CONT");
      econt->ex_out(extract_o);
   icont = new inw_cont("INW_CONT");
      icont->ij_inp(inject_i);
   pn  = new proc_node* [XSIZE * YSIZE];

   for(int n= (XSIZE * YSIZE) -1; n>=0 ; n--) {
      sprintf(nodename, "RT%d", n);
      rt[n] = new con_t(nodename, BYTES_FIFO_DEPTH/BYTES_OHEAD);

      sprintf(nodename, "LT%d", n);
      lt[n] = new con_t(nodename, BYTES_FIFO_DEPTH/BYTES_OHEAD);

      sprintf(nodename, "UP%d", n);
      up[n] = new con_t(nodename, BYTES_FIFO_DEPTH/BYTES_OHEAD);

      sprintf(nodename, "DN%d", n);
      dn[n] = new con_t(nodename, BYTES_FIFO_DEPTH/BYTES_OHEAD);

      sprintf(nodename, "ENW%d", n);
      enw[n] = new con_t(nodename, BYTES_FIFO_DEPTH/BYTES_OHEAD);

      sprintf(nodename, "INW%d", n);
      inw[n] = new con_t(nodename, BYTES_FIFO_DEPTH/BYTES_OHEAD);

// IENW connections
      econt->ex_inp[n](*enw[n]);
      icont->ij_out[n](*inw[n]);
   }

// CNW connections
   for(int i=0, n=0; i< YSIZE; i++)          // i = y coordinate of PN
   for(int j=0; j< XSIZE; j++, n++) {        // j = x coordinate of PN
      sprintf(nodename, "PN%d_%d", j, i);
      pn[n] = new proc_node(nodename, j, i);
         for(int x=0; x<XLEVELS; x++)
            for(int y=0; y<YLEVELS; y++) {
               int xcoord_lev = x*XSIZE + (x%2 ? XSIZE-j-1 : j);
               int ycoord_lev = y*YSIZE + (y%2 ? YSIZE-i-1 : i);
               float data = (gendata(xcoord_lev+0.5, ycoord_lev+0.5) + 0.07) * 1.5; // DJ- Just to keep the path a bit above the landscape

               pn[n]->mapdata[x][y] = data;
            }

         pn[n]->out_port[UP](*up[n]);
         pn[n]->out_port[RIGHT](*rt[n]);
         pn[n]->out_port[DOWN](*dn[n]);
         pn[n]->out_port[LEFT](*lt[n]);

         if(j<XSIZE-1) pn[n]->in_port[RIGHT](*lt[n+1]); // Right
         else          pn[n]->in_port[RIGHT](*rt[n]);

         if(j > 0) pn[n]->in_port[LEFT](*rt[n-1]); // Left
         else      pn[n]->in_port[LEFT](*lt[n]);

         if(i<YSIZE-1) pn[n]->in_port[UP](*dn[n+XSIZE]); // Up
         else          pn[n]->in_port[UP](*up[n]);

         if(i > 0) pn[n]->in_port[DOWN](*up[n-XSIZE]); // Down
         else      pn[n]->in_port[DOWN](*dn[n]);

         pn[n]->in_port[0](*inw[n]);  // Inject
         pn[n]->out_port[0](*enw[n]); // Extract
   }

}

network::~network()
{
   for(int n=(XSIZE * YSIZE) -1; n >=0 ; n--) {
      delete pn[n];

      if(rt[n]->bytes_free() != BYTES_FIFO_DEPTH) cout << "problem in rt[" << n << "]" << endl;
      if(lt[n]->bytes_free() != BYTES_FIFO_DEPTH) cout << "problem in lt[" << n << "]" << endl;
      if(up[n]->bytes_free() != BYTES_FIFO_DEPTH) cout << "problem in up[" << n << "]" << endl;
      if(dn[n]->bytes_free() != BYTES_FIFO_DEPTH) cout << "problem in dn[" << n << "]" << endl;

      delete rt[n];
      delete lt[n];
      delete up[n];
      delete dn[n];
      delete inw[n];
      delete enw[n];
   }

   delete[] pn;
   delete[] rt;
   delete[] lt;
   delete[] up;
   delete[] dn;
   delete[] inw;
   delete[] enw;
  
   delete econt;
   delete icont;

}
