#include <math.h>
#include "proc_node.h"
#include "global.h"


int proc_node::route(int *vias_here)
{
      float exit_y, exit_x;
      local_path_len = 0;
      float new_entry_x, new_entry_y;
      float cur_entry_x, cur_entry_y;
      int route_dir;

 // Effective unfloded x & y coordinates of present processing node.
      int xcoord_lev = (ftr_loc->foldlevelx*XSIZE) + (ftr_loc->foldlevelx%2 ? XSIZE-xcoord-1 : xcoord);
      int ycoord_lev = (ftr_loc->foldlevely*YSIZE) + (ftr_loc->foldlevely%2 ? YSIZE-ycoord-1 : ycoord);

      if(ftr_loc->debug)
         ftr_loc->debfile << ftr_loc->entry_x+(float) xcoord_lev << " " << ftr_loc->entry_y+(float) ycoord_lev << " " << mapdata[ftr_loc->foldlevelx][ftr_loc->foldlevely] << " " << ftr_loc->foldlevelx << " " << ftr_loc->foldlevely << " " << ftr_loc->data << endl;

// Check if the current map segment contains the next via point
// Repeat until a via point outside current segment is found
      *vias_here = 0;

      while(((int) ftr_loc->via_x[0] - xcoord_lev) == 0
      && ((int) ftr_loc->via_y[0] - ycoord_lev) == 0
      && ftr_loc->no_via >0 ) {

         if(ftr_loc->debug)
            ftr_loc->debfile << ftr_loc->via_x[0] << " " << ftr_loc->via_y[0] << " " << mapdata[ftr_loc->foldlevelx][ftr_loc->foldlevely] << " " << ftr_loc->foldlevelx << " " << ftr_loc->foldlevely << " " << ftr_loc->data << " - Via" << endl;

         new_entry_x = ftr_loc->via_x[0] - xcoord_lev;
         new_entry_y = ftr_loc->via_y[0] - ycoord_lev;

         local_path_len +=length(ftr_loc->entry_x, ftr_loc->entry_y, new_entry_x, new_entry_y);

         ftr_loc->entry_x = new_entry_x;
         ftr_loc->entry_y = new_entry_y;

         ftr_loc->no_via--;
         ftr_loc->via_x++;
         ftr_loc->via_y++;
         (*vias_here)++;
      }

      ftr_loc->forward = 0;
      cur_entry_x = ftr_loc->entry_x;
      cur_entry_y = ftr_loc->entry_y;

      float XDIFF = (ftr_loc->via_x[0] - ftr_loc->entry_x - (float) xcoord_lev);
      float YDIFF = (ftr_loc->via_y[0] - ftr_loc->entry_y - (float) ycoord_lev);

      if(ftr_loc->no_via == 0) // To extract
         route_dir = 0;

// CNW Routing calculations
      else {
         (*vias_here)++;

         if(XDIFF == 0)       // Takes care of vertical routing!
            exit_y = YDIFF * INFINITY; //+ cur_entry_y;  // Dummy value

         else if(XDIFF > 0)
            exit_y = (YDIFF/XDIFF)*(1.0-cur_entry_x) + cur_entry_y; // (1, exit_y) is one possible exit point
         else
            exit_y = (YDIFF/XDIFF)*(0.0-cur_entry_x) + cur_entry_y; // (0, exit_y) is one possible exit point


         if(YDIFF == 0)       // Takes care of horizontal routing!
            exit_x = XDIFF * INFINITY; // + cur_entry_x;  // Dummy value.

         else if(YDIFF >= 0)
            exit_x = (XDIFF/YDIFF)*(1.0-cur_entry_y) + cur_entry_x; // (exit_x, 1) is one possible exit point
         else
            exit_x = (XDIFF/YDIFF)*(0.0-cur_entry_y) + cur_entry_x; // (exit_x, 0) is one possible exit point

         if(ftr_loc->debug)
            ftr_loc->debfile << "#xdiff, ydiff, exit_x, exit_y = " << XDIFF << ", " << YDIFF << ", " << exit_x << ", " << exit_y << ": ";

// Direction ditermination
// If doing diagonal routing; disable it when the via point lies in the next node.
// Additional path length due to diagonal considered in the next proc node
         if(exit_y >= 1.0) { // To Up
            ftr_loc->entry_x = exit_x;
            ftr_loc->entry_y = 0.0;
            local_path_len += length(cur_entry_x, cur_entry_y, exit_x, 1.0);
            route_dir = UP;

            if((int) ftr_loc->via_y[0] != ycoord_lev+1 || (int) ftr_loc->via_x[0] != xcoord_lev) { // Next segment does not contain next via
               if(exit_y < 1.0+DIAG_SEP_TH && exit_x < DIAG_SEP_TH && XDIFF < 0) { // Diagonally to up-left; if XDIFF is not checked,
                  ftr_loc->entry_x += 1.0;                   // it will be wrong for paths cutting x-axis at very shallow angles, so that
                  ftr_loc->forward = LEFT;                   // exit_y & exit_x satisfy the condition without being near to the same corner
               }
               else if(exit_y < 1.0+DIAG_SEP_TH && exit_x > 1.0-DIAG_SEP_TH && XDIFF > 0) { // Diagonally to up-right
                  ftr_loc->entry_x -= 1.0;
                  ftr_loc->forward = RIGHT;
               }
            }
         }
         else if(exit_y <= 0.0) { // To Down
            ftr_loc->entry_x = exit_x;
            ftr_loc->entry_y = 1.0;
            local_path_len += length(cur_entry_x, cur_entry_y, exit_x, 0.0);
            route_dir = DOWN;

            if((int) ftr_loc->via_y[0] != ycoord_lev-1 || (int) ftr_loc->via_x[0] != xcoord_lev) { // Next segment does not contain next via
               if(exit_y > -DIAG_SEP_TH && exit_x < DIAG_SEP_TH && XDIFF < 0) { // Diagonally to down-left
                  ftr_loc->entry_x += 1.0;
                  ftr_loc->forward = LEFT;
               }
               else if(exit_y > -DIAG_SEP_TH && exit_x > 1.0-DIAG_SEP_TH && XDIFF > 0) { // Diagonally to down-right
                  ftr_loc->entry_x -= 1.0;
                  ftr_loc->forward = RIGHT;
               }
            }
         }
         else if(exit_x >= 1.0) { // To Right
            ftr_loc->entry_x = 0.0;
            ftr_loc->entry_y = exit_y;
            local_path_len += length(cur_entry_x, cur_entry_y, 1.0, exit_y);
            route_dir = RIGHT;

            if((int) ftr_loc->via_y[0] != ycoord_lev || (int) ftr_loc->via_x[0] != xcoord_lev+1) { // Next segment does not contain next via
               if(exit_y < DIAG_SEP_TH && exit_x < 1.0+DIAG_SEP_TH && YDIFF < 0) { // Diagonally to down-right
                  ftr_loc->entry_y += 1.0;
                  ftr_loc->forward = DOWN;
               }
               else if(exit_y > 1.0-DIAG_SEP_TH && exit_x < 1.0+DIAG_SEP_TH && YDIFF > 0) { // Diagonally to up-right
                  ftr_loc->entry_y -= 1.0;
                  ftr_loc->forward = UP;
               }
            }
         }
         else if(exit_x <= 0.0) { // To Left
            ftr_loc->entry_x = 1.0;
            ftr_loc->entry_y = exit_y;
            local_path_len += length(cur_entry_x, cur_entry_y, 0.0, exit_y);
            route_dir = LEFT;


            if((int) ftr_loc->via_y[0] != ycoord_lev || (int) ftr_loc->via_x[0] != xcoord_lev-1) { // Next segment does not contain next via
               if(exit_y < DIAG_SEP_TH && exit_x > -DIAG_SEP_TH && YDIFF < 0) { // Diagonally to down-left
                  ftr_loc->entry_y += 1.0;
                  ftr_loc->forward = DOWN;
               }
               else if(exit_y > 1.0-DIAG_SEP_TH && exit_x > -DIAG_SEP_TH && YDIFF > 0) { // Diagonally to up-left
                  ftr_loc->entry_y -= 1.0;
                  ftr_loc->forward = UP;
               }
            }
         }
         else
            cout << "Unknown Route in " << xcoord_lev << "," << ycoord_lev << endl; // Should never really happen

      if(ftr_loc->debug)
         ftr_loc->debfile << "route_dir, forward: " << route_dir << ", " << ftr_loc->forward << endl;
      }

// Routing direction adjustment due to folding
      if(ftr_loc->foldlevelx%2) {
         if(route_dir == LEFT) route_dir = RIGHT;
         else if(route_dir == RIGHT) route_dir = LEFT;

         if(ftr_loc->forward == LEFT) ftr_loc->forward = RIGHT;
         else if(ftr_loc->forward == RIGHT) ftr_loc->forward = LEFT;
      }

      if(ftr_loc->foldlevely%2) {
         if(route_dir == UP) route_dir = DOWN;
         else if(route_dir == DOWN) route_dir = UP;

         if(ftr_loc->forward == UP) ftr_loc->forward = DOWN;
         else if(ftr_loc->forward == DOWN) ftr_loc->forward = UP;
      }

// Change of folding level <-> Bouncing of packet at processing array edge
// DJ: Presently using the feedback paths at the processing array edges!!
      if(xcoord == XSIZE-1 && (route_dir == RIGHT || ftr_loc->forward == RIGHT)) {
         if(ftr_loc->foldlevelx%2) ftr_loc->foldlevelx--;
         else ftr_loc->foldlevelx++;
      }
      else if(xcoord == 0  && (route_dir == LEFT || ftr_loc->forward == LEFT)) {
         if(ftr_loc->foldlevelx%2) ftr_loc->foldlevelx++;
         else ftr_loc->foldlevelx--;
      }

      if(ycoord == 0  && (route_dir == DOWN || ftr_loc->forward == DOWN)) {
         if(ftr_loc->foldlevely%2) ftr_loc->foldlevely++;
         else ftr_loc->foldlevely--;
      }
      else if(ycoord == YSIZE-1 && (route_dir == UP || ftr_loc->forward == UP)) {
         if(ftr_loc->foldlevely%2) ftr_loc->foldlevely--;
         else ftr_loc->foldlevely++;
      }

      return route_dir;
}


void proc_node::impl_proc_node() {
   int route_dir;

      while(true) {

         arbit();  // Arbiter logic - select which input to be read, if any

         if(q_next_pos == 1) { // There is no data to be read in the waiting queue
            wait(CLK_PER);
            continue;
         }

// A packet is being processed now: measure the busy time to estimate processor loading
         total_busy_time = total_busy_time - sc_time_stamp();

         read_input(); // Reads the input packet into ftr_loc, but does not remove the packet from FIFO/memory

         send_list[read_from_dir].old_bytes = BYTES_OHEAD + (ftr_loc->no_via)*BYTES_PER_VIA;


         ftr_loc -> proc_cost++; // Incrementing processing cost spent to include the last communication.

         if(DEBUG_NODES)
            cout << "ThereAwt" << xcoord << "," << ycoord << "\t" << ftr_loc->data << endl;

         if(ftr_loc -> forward == 0) {
            int vias_here; // Number of via points to be read for the processing in this proc node

            route_dir = route(&vias_here); // Routing logic - calculates the routing parameters. Does not actually send packet

   // Emulating the fifo-delay
   // Only those vias which are in the present map segment and one via afterwards, if any, are to be read
            wait(CLK_PER*((BYTES_OHEAD + (vias_here*BYTES_PER_VIA) +FIFO_W-1)/FIFO_W)); // Rounding up
            if(ftr_loc->no_via == 0)
               wait(vias_here*VIA_PROC_TIME);
            else
               wait((vias_here-1)*VIA_PROC_TIME + ROUTE_DELAY);

            func();

            wait(PROC_TIME);         // Emulates the data processing time
            ftr_loc -> proc_cost++; // Incrementing processing cost spent to include the last computation.
            this->num_pack_proc++;
         }
         else {
            if(DEBUG_NODES)
               cout << "Forwarding to " << ftr_loc->forward << " in " << xcoord << "," << ycoord << endl;

            route_dir = ftr_loc -> forward;
            ftr_loc -> forward = 0;
            wait(2*CLK_PER); // One clock to read the forward bits, the other to write them back.
         }

// Putting the packet into the send-list
         send_list[read_from_dir].ftr_in_list = ftr_loc; // No actual write physically. Packet already in memory.

         wait(8*CLK_PER); // 1 memory read: no_via, proc_cost; 2 Adds, 2 Mults, 1 not, assignments
         send_list[read_from_dir].bytes = BYTES_OHEAD + (ftr_loc->no_via)*BYTES_PER_VIA;
         send_list[read_from_dir].tout_cnt = (BASE_DL_TOUT+DL_TOUT_COST*ftr_loc->proc_cost)/CLK_PER;
         send_list[read_from_dir].route_dir = route_dir;
         send_list[read_from_dir].ready_toggle = !send_list[read_from_dir].ready_toggle; // Packet ready to be sent

         total_busy_time = total_busy_time + sc_time_stamp();

      }
   }

