#include <systemc.h>
#include "ftr.h"
#include "if.h"
#include "simparam.h"


SC_MODULE (proc_node) {
private:
   struct {
     bool write_toggle, read_toggle, ready_toggle;
     int  route_dir;
     int  tout_cnt;
     int  bytes, old_bytes;   // BYTES_PER_VIA*(ftr_loc->no_via) + BYTES_OHEAD
     ftr_t* ftr_in_list;
   } send_list[5];

   int    xcoord, ycoord;
   int    read_from_dir;
   ftr_t*  ftr_loc;

   float  local_path_len; // Length of the path in current map segment

   int  arbit_q[5]; // Arbitration queue for RR policy
                    // arbit_q[x] == y => x-th port at position y in q. y == 0 => not in q
                    //         x==0...4 => inject, up, right, down, left
   int  q_next_pos; // The position in q where the next incoming packet will enter

   int route(int*);

   float length(float ax, float ay, float bx, float by) {
      return sqrt((by-ay)*(by-ay)+(bx-ax)*(bx-ax));
   } 

   void arbit() { 
      for(int i=0; i < 5; i++) {
         if(arbit_q[i] == 0 && in_port[i]->num_available()>0   // Not in the input queue presently
            && send_list[i].read_toggle == send_list[i].write_toggle) // Not in the send-list presently
         {
            arbit_q[i] = q_next_pos;
            q_next_pos++;
         }
      }
   }

   void read_input() {
      ftr_loc = NULL;
      for(int i=0; i < 5; i++) {
         if(arbit_q[i] == 1) {
            ftr_loc = in_port[i]->read_init();

            q_next_pos--;
            send_list[i].read_toggle = !send_list[i].read_toggle;
            read_from_dir = i;
         }
         if(arbit_q[i] > 0)
            arbit_q[i]--;
      }
   }

   void func() {
         ftr_loc->data += local_path_len;      // Emulates the data processing
   }


   void send_packet() {
      while(true) {
         for(int i=0; i<5; i++) {
            if(send_list[i].ready_toggle != send_list[i].write_toggle) {    // data ready to be sent
               if(out_port[send_list[i].route_dir]->bytes_free() < send_list[i].bytes) { // Not enough space in fifo
                  send_list[i].tout_cnt--;
                  if(send_list[i].tout_cnt == 0)
                     send_list[i].route_dir = 0;  // Deadlock detected
               }
               else {  // Enough space in FIFO to send the packet
                  wait(CLK_PER * ((send_list[i].bytes +FIFO_W-1)/FIFO_W));
                  in_port[i]->read_finalize(send_list[i].old_bytes);  // Finalize the read, i.e. the packet has been fully read out of the fifo
                  wait(CLK_PER);
                  out_port[send_list[i].route_dir]->write(send_list[i].ftr_in_list);
                  send_list[i].write_toggle = !send_list[i].write_toggle; // The packet is removed from the send list
               }
            }
         }

         wait(CLK_PER);
      }
   }

public:

   outp_t out_port[5];  // 0...4 => extract, up, right, down, left
   inp_t  in_port[5];  // 0...4 => insert, up, right, down, left
   float  mapdata[XLEVELS][YLEVELS];

   int    num_pack_proc; // Number of packets processed.
   sc_time total_busy_time;

   void impl_proc_node();

   proc_node (sc_module_name name, int x, int y) : sc_module(name) {
      xcoord = x;
      ycoord = y;

      ftr_loc = NULL;
      num_pack_proc = 0;

      sc_time tmp_time(0, SC_NS);
      total_busy_time = tmp_time;

      arbit_q[0] = arbit_q[1] = arbit_q[2] = arbit_q[3] = arbit_q[4] = 0;
      q_next_pos = 1;

      send_list[0].write_toggle = send_list[0].read_toggle = send_list[0].ready_toggle = false;
      send_list[1].write_toggle = send_list[1].read_toggle = send_list[1].ready_toggle = false;
      send_list[2].write_toggle = send_list[2].read_toggle = send_list[2].ready_toggle = false;
      send_list[3].write_toggle = send_list[3].read_toggle = send_list[3].ready_toggle = false;
      send_list[4].write_toggle = send_list[4].read_toggle = send_list[4].ready_toggle = false;

      SC_HAS_PROCESS(proc_node);

      SC_THREAD(impl_proc_node);
      SC_THREAD(send_packet);
   }
};

