#include <systemc.h>
#include "network.h"
#include "simparam.h"
#include <stdlib.h>
#include <stdio.h>


SC_MODULE(testbench) {
private:
   ofstream plotscript;
   int cnt_dl;
   float max_ppm_error;
   int max_error_pack_id;
   int num_debpack;
   int *debpack;
   unsigned int seed;
   bool pack_received[NUM_PACK];
   int wasted_cost;
   sc_time latency;

   int intermx[2], intermy[2];

   int num_pack_sent, num_pack_rec;

   float length(float ax, float ay, float bx, float by) {
      return sqrt((by-ay)*(by-ay)+(bx-ax)*(bx-ax));
   } 

/*
 // Original test generator
   ftr_t* gen_rand_pack(int id) {
      ftr_t*   test_ftr;
      bool     debug_packet = DEBUG_PACKETS;

      for(int i=0; i<num_debpack; i++)
         if(id==debpack[i]) debug_packet = true;

      int num_via = (rand() % (MAX_NUM_VIA-1))+2; // 2 .. MAX_NUM_VIA
//      int num_via = MAX_NUM_VIA; // Trying to overload the network and create deadlocks

      float via_x = (float) rand() * (float) (XSIZE*XLEVELS) / (float) RAND_MAX;
      float via_y = (float) rand() * (float) (YSIZE*YLEVELS) / (float) RAND_MAX;

// (packet ID, number of via, start x, start y, is debug?)
      test_ftr = new ftr_t(id, num_via, via_x, via_y, debug_packet);
      if(debug_packet) test_ftr->debfile << "#(" << via_x << ", " << via_y << ")" << endl;

      for(int i=1; i<num_via; i++) {
         via_x = (float) rand() * (float) (XSIZE*XLEVELS) / (float) RAND_MAX;
         via_y = (float) rand() * (float) (YSIZE*YLEVELS) / (float) RAND_MAX;
         if(debug_packet) test_ftr->debfile << "#(" << via_x << ", " << via_y << ")" << endl;

         test_ftr->via_x[i]=via_x;
         test_ftr->via_y[i]=via_y;

         test_ftr->expected_len += length(test_ftr->via_x[i-1], test_ftr->via_y[i-1], test_ftr->via_x[i], test_ftr->via_y[i]);
         if(debug_packet) test_ftr->debfile << "#" << test_ftr->expected_len << endl;
      }

      return test_ftr;
   } */


// DJ - rigged deterministic function
   ftr_t* gen_rand_pack(int id) {
      ftr_t*   test_ftr;
      static int intx1 = rand() % (XSIZE * XLEVELS);
      static int inty1 = rand() % (YSIZE * YLEVELS);
      static int intx2 = rand() % (XSIZE * XLEVELS);
      static int inty2 = rand() % (YSIZE * YLEVELS);
      int num_via = 4; // 2 .. MAX_NUM_VIA

      float via_x = 8.5;
      float via_y = 5.5;
 /*
      intermx[0] = intx1;
      intermy[0] = inty1;
      intermx[1] = intx2;
      intermy[1] = inty2; */

// (packet ID, number of via, start x, start y, is debug?)
      test_ftr = new ftr_t(id, num_via, via_x, via_y, false);
/*
      for(int i=1; i<num_via; i++) {
         via_x = 18.0813;
         via_y =  4.1041;

         test_ftr->via_x[i]=via_x;
         test_ftr->via_y[i]=via_y;

         test_ftr->expected_len += length(test_ftr->via_x[i-1], test_ftr->via_y[i-1], test_ftr->via_x[i], test_ftr->via_y[i]);
      } */
/*
      test_ftr->via_x[1] = intx1 + rand()/((float) RAND_MAX);
      test_ftr->via_y[1] = inty1 + rand()/((float) RAND_MAX);
      test_ftr->via_x[2] = intx2 + rand()/((float) RAND_MAX);
      test_ftr->via_y[2] = inty2 + rand()/((float) RAND_MAX); */

      test_ftr->via_x[1] = rand()*((XSIZE*XLEVELS)/((float) RAND_MAX));
      test_ftr->via_y[1] = rand()*((YSIZE*YLEVELS)/((float) RAND_MAX));
      test_ftr->via_x[2] = rand()*((XSIZE*XLEVELS)/((float) RAND_MAX));
      test_ftr->via_y[2] = rand()*((YSIZE*YLEVELS)/((float) RAND_MAX));

      test_ftr->via_x[3] = 181.5;
      test_ftr->via_y[3] = 188.2;

      test_ftr->expected_len += length(test_ftr->via_x[0], test_ftr->via_y[0], test_ftr->via_x[1], test_ftr->via_y[1]);
      test_ftr->expected_len += length(test_ftr->via_x[1], test_ftr->via_y[1], test_ftr->via_x[2], test_ftr->via_y[2]);

      return test_ftr;
   }


void print_results() {

   cout << endl << endl << "Number of packets injected = " << NUM_PACK << endl;
   cout << "Number of packets ejected due to deadlock = " << cnt_dl << endl;
   cout << "Packet injection-extraction throughput = " << THROUGHPUT << " clocks/packet" << endl;
   cout << "Maximum relative error in calculation = " << max_ppm_error << " ppm in packet ID: " << max_error_pack_id << endl;
   cout << "Simulation seed: " << seed << endl;
   cout << "Wasted resources due to deadlock (computation = communication = unit resource): " << wasted_cost << endl;
   cout << "Average proc node loading = " << nw1->get_network_load()*100 << "% ... More details in network_load.log" << endl;
   cout << "Average latency of non-timed-out packets = " << (latency /(NUM_PACK-cnt_dl)) / CLK_PER << " clock periods" << endl;
   cout << "Intermediate map segements: (" << intermx[0] << ", " << intermy[0] << "), (" << intermx[1] << ", " << intermy[1] << ")" << endl;
}


void run_tb_prod() {
ftr_t* pack;

   for(int i=0; i<NUM_PACK; i++) {
      pack=gen_rand_pack(i);    // Assigning the storage for a packet

      inject->write(pack);
      cout << "Packet ID: " << i << " launched at time: " << sc_time_stamp() << endl;
      num_pack_sent++;
      wait(CLK_PER * THROUGHPUT);
   }
}

void run_tb_cons() {
int cnt = 0;
FILE *pl = fopen("act_packets.log", "w"); // Log of number of active packets
ftr_t *ext;
sc_time last_time;
bool flag_interim = false;
float error;
bool flag=false;

plotscript << "set xtics 0, 1, " << XSIZE*XLEVELS << endl;
plotscript << "set ytics 0, 1, " << YSIZE*YLEVELS << endl;
plotscript << "set xrange[0:" << XSIZE*XLEVELS << "]" << endl;
plotscript << "set yrange[0:" << YSIZE*YLEVELS << "]" << endl;
plotscript << "set grid ytics lc rgb \"#bbbbbb\" lw 1 lt 0" << endl;
plotscript << "set grid xtics lc rgb \"#bbbbbb\" lw 1 lt 0" << endl << endl;

plotscript << "plot ";

fprintf(pl, "Clock cycles\tNum of Active packets\tNum of packets ejected\n");

do {
   while(extract->num_available()>0) {
      ext = extract->read();
      cnt++;
      pack_received[ext->ID] = true;
      last_time = sc_time_stamp();
      fprintf(pl, "%d\t%d\t%d\n", (int) (last_time/CLK_PER), num_pack_sent-cnt, ext->no_via != 0 ? 1 : 0);
      flag_interim = false;

      if(ext->no_via != 0) {    // Packet calculation was not finished, i.e. ejected due to deadlock time-out
         cout << "Packet ID: " << ext->ID << " ejected because of deadlock at time: " << sc_time_stamp() << endl;
         cnt_dl++;
         wasted_cost += ext->proc_cost;
      }
      else {
         latency += (ext->out_time - ext->in_time);
         error = ext->data - ext->expected_len;
         cout << "Packet ID: " << ext->ID << " finished at time: " << sc_time_stamp() << "; Expected length = " << ext->expected_len << "; Error = " << error <<" ; " << cnt << " packets finished.";
         error /= ext->expected_len;
         error *= 1000000;
         if(error < 0) error = -error;
         if(error >= max_ppm_error) {
            max_ppm_error = error;
            max_error_pack_id = ext->ID;
         }
         cout << endl;
      }

      if(ext->debug) {
         if(flag) plotscript << ", ";
         flag = true;
         plotscript << "\\" << endl << "\"" << ext->debfilename << "\" using 1:2 with linespoints";
      }

      delete ext;  // Freeing the storage for the packet
   }
   wait(CLK_PER);

// Enable printing intermediate results if the last few packets are too delayed!
// -- They might be lost due to sim environment faults! Indeed happening once in a while!
   if(cnt > (NUM_PACK*199)/200 && flag_interim == false && sc_time_stamp()-last_time > 100*PROC_TIME) {
      flag_interim = true;
      cout << "Seems like some packet got lost! Interim results:" << endl;
      print_results();
      cout << "Probably lost packet IDs: ";
      for(int i=0; i<NUM_PACK; i++)
         if(pack_received[i]==false) cout << i << " ";
      cout << endl;
   }

} while (cnt < NUM_PACK);

plotscript << endl << endl << "pause -1" << endl;
plotscript.close();

print_results();
fclose(pl);

if(flag) // i.e. at least some debug packet was passed
   system("gnuplot ./plotscript.gp");

sc_stop();
}


public:
   con_t *inject;
   con_t *extract;
   network *nw1;


testbench(sc_module_name name, unsigned int seed_, int num_debpack_, int *debpack_) : sc_module(name) {

   cnt_dl = 0;
   max_ppm_error = 0;
   max_error_pack_id = -1;

   num_debpack = num_debpack_;
   debpack = debpack_;
   seed = seed_;
   wasted_cost = 0;

   sc_time tmp_time(0, SC_NS);
   latency = tmp_time;
   num_pack_sent = num_pack_rec = 0;

   inject = new con_t("MAIN_INJ", 4); // DJ: Depth 4 packets - should be ok
   extract = new con_t("MAIN_EXT", 4); // DJ: Depth 4 packets - should be ok

   nw1 = new network("NW");
      nw1->inject_i(*inject);
      nw1->extract_o(*extract);

   srand(seed_);

   plotscript.open("plotscript.gp");
   plotscript << "#Simulation seed = " << seed_ << endl << endl;

   for(int i=0; i<NUM_PACK; i++)
      pack_received[i] = false;
   
   SC_HAS_PROCESS(testbench);

   SC_THREAD(run_tb_prod);
   SC_THREAD(run_tb_cons);
}

~testbench() {
   cout << "Data structures deleted" << endl;

   delete inject;
   delete extract;
   delete nw1;
}

};
